/*
 *  Copyright (C) 2005 Steve Harris
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  $Id: assert.c,v 1.12 2005/11/16 18:46:51 theno23 Exp $
 */

#include <stdio.h>
#include <string.h>
#include <glib.h>
#include <mysql.h>
#include <mysqld_error.h>
#include <sys/types.h>
#include <time.h>
#include <sys/time.h>

#include "tstore.h"
#include "import.h"
#include "timing.h"
#include "rdf_constants.h"
#include "config.h"

#if MYSQL_VERSION_ID >= 40100
#define SQL_TRUE  "true"
#define SQL_FALSE "false"
#else
#define SQL_TRUE  "1"
#define SQL_FALSE "0"
#endif

#define TMP_SIZE 256

#define HASH_SIZE 4096

typedef struct ts_raw_triple_s {
        ts_id modl;
        ts_id subj;
        ts_id pred;
        ts_id obj;
} ts_raw_triple;

static int ts_block_append(ts_connection *db, ts_raw_triple *block, int *count,
                            ts_hash m, ts_hash s, ts_hash p, ts_hash o);

static void ts_block_flush(ts_connection *db, ts_raw_triple *block, int *count);

/* run the inserts from the symbol vector */
//static void ts_block_ensure_resource_flush(ts_connection *c);

#define BLOCK_SIZE 4096
#define URI_BLOCK_SIZE 1024

static int ts_tb_count = 0;
static ts_raw_triple ts_pending_block[BLOCK_SIZE];

#if 0
static int ts_rb_count = 0;
static char *ts_symbol_block[URI_BLOCK_SIZE];
#endif

int ts_assert_triple(ts_connection *c, ts_hash model, const char *subj,
			  const char *pred, const char *obj,
			  const ts_obj_type type, int dt, int lang)
{
    ts_hash s = 0, p = 0, o = 0;

    if (type == ObjAny) {
	return TS_BAD_TYPE;
    }

    s = TS_ID(c, ts_ensure_resource(c, subj));
    p = TS_ID(c, ts_ensure_resource(c, pred));
    if (type == ObjURI) {
	o = TS_ID(c, ts_ensure_resource(c, obj));
    } else {
	o = TS_ID(c, ts_ensure_literal(c, (char *)obj, dt, lang));
    }

    if (c->block_mode) {
	return ts_block_append(c, ts_pending_block, &ts_tb_count,
			       TS_ID(c, model), s, p, o);
    } else {
	return ts_assert_raw(c, TS_ID(c, model), s, p, o);
    }
}

#if 0
/* only works in triple mode, but it blocks up writes to the symbols table */
ts_hash ts_block_ensure_resource(ts_connection *c, const char *uri)
{
    MYSQL_RES *res;
    MYSQL_ROW row;
    char tmp[TMP_SIZE];
    char *buffer;
    ts_hash hash;
    uint32_t id;
    static GHashTable *recent = NULL;
    static uint32_t id_num = 0xFFFFFFFF;
    static uint32_t id_count = 0;
    uint32_t retval;
    char *query;

TIME(NULL);

    if (!recent) {
	recent = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, NULL);
    }
    if (!uri) {
	/* this signals that the symbols table has been cleared */
	g_hash_table_destroy(recent);
	recent = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, NULL);

	return 0;
    }

    if ((id = (uint32_t)g_hash_table_lookup(recent, uri))) {
TIME("hash_hit");
	return id;
    }
    hash = ts_hash_uri(uri);
    snprintf(tmp, TMP_SIZE-1, "SELECT sid FROM symbols WHERE hash=%lld LIMIT 1",
	     hash);
    ts_mysql_query(c->db, tmp);
    res = mysql_use_result(c->db);
    if (res && (row = mysql_fetch_row(res))) {
	id = atoll(row[0]);
	mysql_free_result(res);
	g_hash_table_insert(recent, strdup(uri), (gpointer)id);
TIME("sql_hit");

	return id;
    }
    mysql_free_result(res);

    if (id_count-- == 0) {
	id_count = 1024;
	id_num = ts_claim_ai_vals(c, "symbols", "sid", id_count);
    }
    
    buffer = malloc(strlen(uri) * 2 + 1);
    mysql_real_escape_string(c->db, buffer, uri, strlen(uri));
    query = g_strdup_printf("(%d, %lld, '%s', " "NULL, NULL, NULL, NULL, "
				   "%d, 0)", id_num, hash, buffer, c->uri_dt);
    ts_symbol_block[ts_rb_count++] = query;
    ts_symbol_block[ts_rb_count] = NULL;
    free(buffer);

    retval = id_num++;

    if (ts_rb_count == URI_BLOCK_SIZE - 1) {
	ts_block_ensure_resource_flush(c);
    }
TIME("block_add");

    return retval;
}

static void ts_block_ensure_resource_flush(ts_connection *c)
{
    char *inserts;
    char *query;

    if (ts_rb_count == 0) {
	return;
    }

    inserts = g_strjoinv(",", ts_symbol_block);
    query = g_strdup_printf("INSERT IGNORE INTO "T_S" VALUES%s", inserts);
    ts_mysql_query(c->db, query);
    free(inserts);
    free(query);
    ts_rb_count = 0;
    ts_symbol_block[0] = NULL;
}
#endif

int ts_assert_raw(ts_connection *c, ts_hash model, ts_hash subj, ts_hash pred,
		  ts_hash obj)
{
    char tmp[TMP_SIZE];
#ifndef ENABLE_QUADS
    MYSQL_RES *res;
    MYSQL_ROW row;

    snprintf(tmp, TMP_SIZE,
	     "SELECT tid FROM "T_T" WHERE subject=%lld AND "
	     "predicate=%lld AND object=%lld LIMIT 1", 
	     subj, pred, obj);
    ts_mysql_query(c->db, tmp);
    res = mysql_store_result(c->db);
    row = mysql_fetch_row(res);

    /* If the triple allready exists then just add a reference to it */
    if (row) {
	snprintf(tmp, TMP_SIZE, "INSERT IGNORE INTO "T_M" "
		 "VALUES(%s, %lld)", row[0], model);
	if (ts_mysql_query(c->db, tmp)) {
	    return TS_BAD_SQL;
	}
	mysql_free_result(res);

	return 0;
    }
    mysql_free_result(res);
    
    snprintf(tmp, TMP_SIZE,
	     "INSERT INTO "T_T" VALUES(0, %lld, %lld, %lld)",
	    subj, pred, obj);
    if (mysql_query(c->db, tmp)) {
	fprintf(stderr, "SQL error in '%s' %d: %s\n", tmp,
		mysql_errno(c->db), mysql_error(c->db));
	return TS_BAD_SQL;
    }

    snprintf(tmp, TMP_SIZE, "INSERT INTO "T_M" VALUES(%lld, %lld)",
	     mysql_insert_id(c->db), model);
    ts_mysql_query(c->db, tmp);
#else
    snprintf(tmp, TMP_SIZE,
	     "INSERT INTO "T_T" VALUES(%lld, %lld, %lld, %lld)",
	    model, subj, pred, obj);
    if (mysql_query(c->db, tmp)) {
	fprintf(stderr, "SQL error in '%s' %d: %s\n", tmp,
		mysql_errno(c->db), mysql_error(c->db));
	return TS_BAD_SQL;
    }
#endif

    return 0;
}

ts_hash ts_ensure_resource(ts_connection *c, const char *uri)
{
    ts_hash hash;
    static ts_hash recent[HASH_SIZE];
    char *buffer;
    char *query;
#ifndef ENABLE_QUADS
    MYSQL_RES *res;
    MYSQL_ROW row;
#endif
    char tmp[TMP_SIZE];

    if (!uri) {
	/* this signals that the symbols table has been cleared */
	memset(recent, 0, sizeof(recent));

	return 0;
    }

    hash = ts_hash_uri(uri);
    if (recent[hash & (HASH_SIZE - 1)] == hash) {
	return hash;
    }

    recent[hash & (HASH_SIZE - 1)] = hash;

#ifndef ENABLE_QUADS
    snprintf(tmp, TMP_SIZE-1, "SELECT sid FROM "T_S" WHERE hash=%lld LIMIT 1", hash);
    ts_mysql_query(c->db, tmp);
    res = mysql_use_result(c->db);
    row = mysql_fetch_row(res);
    if (row) {
	ts_symbol_cache_id(hash, atoi(row[0]));
	mysql_free_result(res);

	return hash;
    }
    mysql_free_result(res);
#else
    snprintf(tmp, TMP_SIZE-1, "SELECT 1 FROM "T_S" WHERE hash=%lld LIMIT 1", hash);
    if (ts_sql_count(c->db, tmp)) {
	return hash;
    }
#endif

    buffer = malloc(strlen(uri) * 2 + 1);
    mysql_real_escape_string(c->db, buffer, uri, strlen(uri));
#ifndef ENABLE_QUADS
    query = g_strdup_printf("INSERT IGNORE INTO "T_S" VALUES(0, %lld, '%s', "
			    "%d, 0)", hash, buffer, c->uri_dt);
#else
    query = g_strdup_printf("INSERT IGNORE INTO "T_S" VALUES(%lld, '%s', "
			    "NULL, %d, 0)", hash, buffer, c->uri_dt);
#endif
    ts_mysql_query(c->db, query);
    free(query);
#ifndef ENABLE_QUADS
    /* push the ID into the ID cache, so that we can save a query if the next
     * thing is to query for this ID */
    ts_symbol_cache_id(hash, mysql_insert_id(c->db));
#endif

    return hash;
}

#if 1
//XXX LOOSE
static void string_looks_like(const char *str, int *date, int *time)
{
    int len = strlen(str);

    *date = 0;
    *time = 0;

    if (strspn(str, "0123456789-") == len) {
	*date = 1;
    } else if (strspn(str, "0123456789:") == len) {
	*time = 1;
    } else if (strspn(str, "0123456789:-TZ") == len) {
	*time = 1;
	*date = 1;
    }
}
#endif

char *ts_insert_symbol_values(ts_connection *c, GSList **l, ts_hash hash, char *lex, int dt, int lang)
{
    char *buffer;
    char *floating = "NULL", *integer = "NULL", *datetime = "NULL",
	 *ebv = SQL_TRUE;
    char *tmp;
    char timedt[512];

    /* XXX LOOSE check if the lexical form looks like a number */
    char *cnvend;
    int lldate, lltime;
    double val;

    buffer = malloc(strlen(lex) * 2 + 3);
    buffer[0] = '\'';
    mysql_real_escape_string(c->db, buffer+1, lex, strlen(lex));
    strcat(buffer, "'");

    val = strtod(lex, &cnvend);
    if ((lex != cnvend && dt == 0) ||
	dt == c->int_dt ||
	dt == c->decimal_dt ||
	dt == c->float_dt ||
	dt == c->double_dt) {
	floating = buffer;
	integer = buffer;
	if (val != 0.0 && val != -0.0) {
	    ebv = SQL_TRUE;
	} else {
	    ebv = SQL_FALSE;
	}
    }

    if (!strcmp(lex, "false")) {
	ebv = SQL_FALSE;
    }
    
    string_looks_like(lex, &lldate, &lltime);

    if (dt == c->int_dt || dt == c->decimal_dt || dt == c->float_dt ||
	dt == c->double_dt) {
	/* EBV allready calculated */
    } else if (dt == c->datetime_dt) {
	ebv = SQL_TRUE;
    } else if (!strcmp(lex, "true")) {
	ebv = SQL_TRUE;
    } else if (!strcmp(lex, "") || !strcmp(lex, "false")) {
	ebv = SQL_FALSE;
    }

    datetime = ts_normalise_date(c, dt, lex, timedt);
   
    if (strcmp(integer, "NULL")) {
	tmp = g_strdup_printf("INSERT IGNORE INTO "T_I" VALUES(%lld, %s)", hash, integer);
	if (!l) {
	    ts_mysql_query(c->db, tmp);
	    free(tmp);
	} else {
	    *l = g_slist_prepend(*l, tmp);
	}
    }
    if (strcmp(floating, "NULL")) {
	tmp = g_strdup_printf("INSERT IGNORE INTO "T_F" VALUES(%lld, %s)", hash, floating);
	if (!l) {
	    ts_mysql_query(c->db, tmp);
	    free(tmp);
	} else {
	    *l = g_slist_prepend(*l, tmp);
	}
    }
    if (strcmp(datetime, "NULL")) {
	tmp = g_strdup_printf("INSERT IGNORE INTO "T_A" VALUES(%lld, '%s')", hash, datetime);
	if (!l) {
	    ts_mysql_query(c->db, tmp);
	    free(tmp);
	} else {
	    *l = g_slist_prepend(*l, tmp);
	}
    }

    return ebv;
}

ts_hash ts_ensure_literal(ts_connection *c, char *lex, int dt, int lang)
{
#ifndef ENABLE_QUADS
    MYSQL_RES *res;
    MYSQL_ROW row;
#endif
    ts_hash hash;
    static ts_hash recent[HASH_SIZE];
    char *buffer, *ebv;
    char *tmp;
    char query[TMP_SIZE];

    if (!lex) {
	/* this signals that the symbols table has been cleared */
	memset(recent, 0, sizeof(recent));

	return 0;
    }

    hash = ts_hash_literal(lex, dt, lang);
    if (recent[hash & (HASH_SIZE - 1)] == hash) {
	return hash;
    }

#ifndef ENABLE_QUADS
    snprintf(query, TMP_SIZE-1, "SELECT sid FROM "T_S" WHERE hash=%lld LIMIT 1", hash);
    ts_mysql_query(c->db, query);
    res = mysql_use_result(c->db);
    row = mysql_fetch_row(res);
    if (row) {
	ts_symbol_cache_id(hash, atoi(row[0]));
	mysql_free_result(res);

	return hash;
    }
    mysql_free_result(res);
#endif

    recent[hash & (HASH_SIZE - 1)] = hash;

    buffer = malloc(strlen(lex) * 2 + 3);
    buffer[0] = '\'';
    mysql_real_escape_string(c->db, buffer+1, lex, strlen(lex));
    strcat(buffer, "'");

    snprintf(query, TMP_SIZE-1, "SELECT 1 FROM "T_S" where hash=%lld LIMIT 1",
	     hash);
    /* make sure we dont assert the same symbol twice */
    if (ts_sql_count(c->db, query) == 0) {
	ebv = ts_insert_symbol_values(c, NULL, hash, lex, dt, lang);

#ifdef ENABLE_QUADS 
	tmp = g_strdup_printf("INSERT INTO "T_S" VALUES(%lld, %s, %s, %d, %d)",
			      hash, buffer, ebv, dt, lang);
#else
	tmp = g_strdup_printf("INSERT INTO "T_S" VALUES(0, %lld, %s, %d, %d)",
			      hash, buffer, dt, lang);
#endif
	ts_mysql_query(c->db, tmp);
	free(tmp);
    }

#ifndef ENABLE_QUADS
    /* push the ID into the ID cache, so that we can save a query if the next
     * thing is to query for this ID */
    ts_symbol_cache_id(hash, mysql_insert_id(c->db));
#endif

    return hash;
}

/* takes a date format string, and returns a normalised datestring in UTC if
 * possible. The timedt argument should point to a 512 char buffer that may get
 * scribbled on to store the result  */
char *ts_normalise_date(ts_connection *c, int dt, char *buffer, char *timedt)
{
    if (dt == c->date_dt || dt == c->datetime_dt) {
	struct tm t;
	unsigned int dy, dm, dd, th, tm, ts, oh, om;
	char dir;
	int form = sscanf(buffer, "%4u-%2u-%2uT%2u:%2u:%2u%c%2u:%2u", &dy, &dm, &dd, &th, &tm, &ts, &dir, &oh, &om);
	/* If theres a timezone offset then correct for it, to keep internal
	 * values in UTC */
	strncpy(timedt, buffer, 511);
	if (form == 9) {
	    char tmp[512];
	    t.tm_sec = ts;
	    t.tm_min = tm;
	    t.tm_hour = th;
	    t.tm_mday = dd;
	    t.tm_mon = dm - 1;
	    t.tm_year = dy - 1900;

	    strftime(tmp, 512, "%Y-%m-%d %H:%M:%S", &t);
	    snprintf(timedt, 512, "%sDATE(%s, INTERVAL %d MINUTE)",
		    dir == '+' ? "SUB" : "ADD", tmp, oh * 60 + om);
	    return timedt;
	} else {
	    char *zpos = strrchr(timedt, 'Z');
	    
	    /* chop any potential timezone Z off them end, we're in UTC
	     * internally, so can ignore it */
	    if (zpos == timedt + strlen(timedt) - 2) {
		*zpos = '\'';
		*(zpos+1) = '\0';

		return  timedt;
	    }
	}
    } else if (dt == c->time_dt) {
	snprintf(timedt, 512, "'0000-00-00T%s", buffer+1);

	return timedt;
    }

    return "NULL";
}

#if 0
static void dump_pending(MYSQL *mdb, char *t, char *label)
{
    MYSQL_RES *res;
    MYSQL_ROW row;
    char *q;

    printf("%s:\n", label);
    q = g_strdup_printf("SELECT * FROM %s", t);
    ts_mysql_query(mdb, q);
    res = mysql_store_result(mdb);
    printf("s\tp\to\tm\tt\n");
    while ((row = mysql_fetch_row(res))) {
	printf("%s\t%s\t%s\t%s\t%s\n", row[0], row[1], row[2], row[3], row[4]);
    }
    
    printf("\n");
}
#endif

void ts_assert_block_mode(ts_connection *db, int action)
{
    switch (action) {
    case TS_BLOCK_MODE_START:
	if (db->block_mode) {
	    fprintf(stderr, "error: allreading started block mode\n");
	    break;
	}
	db->block_mode = 1;
#ifdef ENABLE_QUADS
#define IDS "bigint"
#else
#define IDS "int"
#endif
	ts_mysql_query(db->db, "CREATE TEMPORARY TABLE pending "
			"(s "IDS" NOT NULL, p "IDS" NOT NULL, "
			 "o "IDS" NOT NULL, m "IDS" NOT NULL, t "IDS" NOT NULL"
			//", KEY(s), KEY(p), KEY(o), KEY(m), KEY(t)"
	")");
	break;
    case TS_BLOCK_MODE_COMMIT_INF:
#ifndef ENABLE_QUADS
	TIME(NULL);
	ts_mysql_query(db->db, "UPDATE pending_i, "T_T" SET t=tid WHERE "
		       "subject=s AND predicate=p AND object=o");
	ts_mysql_query(db->db, "INSERT IGNORE INTO "T_M" SELECT DISTINCT t, m "
		       " FROM pending_i WHERE t != 0");
	ts_mysql_query(db->db, "DELETE FROM pending_i WHERE t != 0");
	ts_mysql_query(db->db, "INSERT IGNORE INTO "T_T" "
			"SELECT DISTINCT 0, s, p, o FROM pending_i");
	ts_mysql_query(db->db, "UPDATE pending_i, "T_T" SET t=tid WHERE "
		       "subject=s AND predicate=p AND object=o");
	ts_mysql_query(db->db, "INSERT IGNORE INTO "T_M" SELECT DISTINCT t, m "
		       " FROM pending_i WHERE t != 0");
	ts_mysql_query(db->db, "UNLOCK TABLES");
	ts_mysql_query(db->db, "DROP TABLE pending_i");
	TIME("commit_inferred_triples");
#endif
	break;
    case TS_BLOCK_MODE_FLUSH:
	ts_block_flush(db, ts_pending_block, &ts_tb_count);
	break;
    case TS_BLOCK_MODE_COMMIT:
	/* dump pending triples into main tables */
	if (!db->block_mode) {
	    fprintf(stderr, "error: not started block mode\n");
	    break;
	}
	TIME(NULL);
	//print_pending(db->db, "pending", "start");
#ifdef ENABLE_QUADS
	ts_mysql_query(db->db, "INSERT IGNORE INTO "T_T" "
			"SELECT DISTINCT m, s, p, o FROM pending");
#else
	ts_mysql_query(db->db, "UPDATE pending, "T_T" SET t=tid WHERE "
		       "subject=s AND predicate=p AND object=o");
	ts_mysql_query(db->db, "INSERT IGNORE INTO "T_M" SELECT DISTINCT t, m "
		       " FROM pending WHERE t != 0");
	ts_mysql_query(db->db, "DELETE FROM pending WHERE t != 0");
	ts_mysql_query(db->db, "INSERT IGNORE INTO "T_T" "
			"SELECT DISTINCT 0, s, p, o FROM pending");
	ts_mysql_query(db->db, "UPDATE pending, "T_T" SET t=tid WHERE "
		       "subject=s AND predicate=p AND object=o");
	ts_mysql_query(db->db, "INSERT IGNORE INTO "T_M" SELECT DISTINCT t, m "
		       " FROM pending WHERE t != 0");
	ts_mysql_query(db->db, "UNLOCK TABLES");
	ts_block_ensure_resource_flush(db);
#endif
	ts_mysql_query(db->db, "DROP TABLE pending");
	db->block_mode = 0;
	TIME("commit_ground_triples");
	break;
    default:
	fprintf(stderr, "Unknown action %d in ts_assert_block_mode()\n",
	        action);
	break;
    }
}

int ts_assert_raw_block(ts_connection *this, ts_raw_triple *block, int count)
{
    char *query;
    int ofs;
    int length;
    int i;

    TIME(NULL);
    length = 100 + 110 * count;
    query = malloc(length * sizeof(char));
    ofs = sprintf(query, "INSERT INTO pending VALUES ");
    for (i = 0; i < count; i++) {
        ofs += snprintf(query + ofs, 110, "("IDFS","IDFS","IDFS","IDFS",0)%s",
                     block[i].subj, block[i].pred, block[i].obj, block[i].modl,
		     i == count - 1 ? "" : ",");
    }
    if (ofs >= length) {
        fprintf(stderr, "Buffer overrun in %s:%d\n", __FILE__, __LINE__);
        return 1;
    }

    ts_mysql_query(this->db, query);
    free(query);
    TIME("block_insert");

    return 0;
}

static void ts_block_flush(ts_connection *db, ts_raw_triple *block, int *count)
{
    if (*count > 0) {
	if (ts_assert_raw_block(db, block, *count)) {
	    fprintf(stderr, "Fatal error, exiting\n");
	    exit(1);
	}
    }
    *count = 0;
}

static int ts_block_append(ts_connection *db, ts_raw_triple *block, int *count,
                            ts_hash m, ts_hash s, ts_hash p, ts_hash o)
{
    block[*count].modl = m;
    block[*count].subj = s;
    block[*count].pred = p;
    block[*count].obj = o;
    if (++*count == BLOCK_SIZE) {
	ts_block_flush(db, block, count);
    }

    return 0;
}

/* vi:set ts=8 sts=4 sw=4: */
