/*
 *  Copyright (C) 2005 Steve Harris
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  $Id: connection.c,v 1.17 2005/10/14 10:07:49 theno23 Exp $
 */

#include <stdlib.h>
#include <string.h>
#include <mysql.h>
#include <rasqal.h>

#include "config.h"
#include "tstore.h"

static int ts_inited = 0;

void ts_init()
{
    ts_inited = 1;
    rasqal_init();
}

void ts_finish()
{
    ts_inited = 0;
    rasqal_finish();
}

ts_connection *ts_connect(const char *host, const char *db,
                          const char *user, const char *passwd)
{
    ts_connection *c = calloc(1, sizeof(ts_connection));
    char *mysql_db;
    char tmp[TS_CF_VAL_SIZE];
    int table_version = 0;

    if (!ts_inited) {
	fprintf(stderr, "fatal error: lib3store not initialised\n");
	return NULL;
    }

    if (!(c->db = mysql_init(NULL))) {
        fprintf(stderr, "Could not initialise database connection\n");

        return NULL;
    }

    mysql_db = g_strdup_printf("%s%s", DBPREFIX, db);
    mysql_options(c->db, MYSQL_READ_DEFAULT_GROUP, "3store");
    if (!mysql_real_connect(c->db, host, user, passwd, mysql_db, 0, NULL, 0)) {
        fprintf(stderr, "Failed to connect to database: Error: %s\n",
                mysql_error(c->db));
	free(mysql_db);

        return NULL;
    }
    c->db_name = mysql_db;

    table_version = ts_config_get_table_version(c->db);
#ifdef ENABLE_QUADS
    if (table_version >= 1000) {
	fprintf(stderr, "fatal error: lib3store built for quad support, but this KB uses triples\n");
	return NULL;
    }
#else
    if (table_version < 1000 && table_version > 0) {
	fprintf(stderr, "fatal error: lib3store built for triple support, but this KB uses quads\n");
	return NULL;
    }
#endif
    if (table_version < TS_CURRENT_TABLE_VERSION) {
	fprintf(stderr, "warning: KB uses old schema, it is recommended to upgrade with\nts-setup -d '%s'\n", db);
    }

#ifdef ENABLE_QUADS
    c->uri_dt = ts_ensure_datatype(c, "_URI");
#else
    c->uri_dt = TS_ID(c, ts_ensure_resource(c, "_URI"));
#endif
    if (!c->uri_dt) {
	fprintf(stderr, "Bad database connection or table setup.\n"
		        "Have you run ts-setup for this database?\n");

	return NULL;
    }

    ts_config_get_value(c, "features", tmp);
    if (strchr(tmp, 'S')) {
	c->feature[TS_FEATURE_STD_COMP] = 1;
    }
    if (strchr(tmp, 'I')) {
	c->feature[TS_FEATURE_INFERENCE] = 1;
    }
    if (strchr(tmp, 'P')) {
	c->feature[TS_FEATURE_PROVENANCE] = 1;
    }
    if (strchr(tmp, 'R')) {
	c->feature[TS_FEATURE_LOADING] = 1;
    }

#ifdef ENABLE_QUADS
    c->int_dt = ts_ensure_datatype(c, XSD_INTEGER);
    c->decimal_dt = ts_ensure_datatype(c, XSD_DECIMAL);
    c->string_dt = ts_ensure_datatype(c, XSD_STRING);
    c->float_dt = ts_ensure_datatype(c, XSD_FLOAT);
    c->double_dt = ts_ensure_datatype(c, XSD_DOUBLE);
    c->datetime_dt = ts_ensure_datatype(c, XSD_DATETIME);
    c->date_dt = ts_ensure_datatype(c, XSD_DATE);
    c->time_dt = ts_ensure_datatype(c, XSD_TIME);
    c->bool_dt = ts_ensure_datatype(c, XSD_BOOLEAN);
#else
    c->int_dt = TS_ID(c, ts_ensure_resource(c, XSD_INTEGER));
    c->decimal_dt = TS_ID(c, ts_ensure_resource(c, XSD_DECIMAL));
    c->string_dt = TS_ID(c, ts_ensure_resource(c, XSD_STRING));
    c->float_dt = TS_ID(c, ts_ensure_resource(c, XSD_FLOAT));
    c->double_dt = TS_ID(c, ts_ensure_resource(c, XSD_DOUBLE));
    c->datetime_dt = TS_ID(c, ts_ensure_resource(c, XSD_DATETIME));
    c->date_dt = TS_ID(c, ts_ensure_resource(c, XSD_DATE));
    c->time_dt = TS_ID(c, ts_ensure_resource(c, XSD_TIME));
    c->bool_dt = TS_ID(c, ts_ensure_resource(c, XSD_BOOLEAN));
#endif

    if (c->uri_dt != 1) {
	fprintf(stderr, "warning: URI datatype is %d not 1, queries may give odd results\n", c->uri_dt);
    }

    c->opt_lvl = DEFAULT_OPT_LVL; /* set default optimisation level */
    c->lock_tables = DEFAULT_LOCK_TABLES;

    mysql_query(c->db, "set collation_connection='utf8_general_ci'");
    mysql_query(c->db, "set collation_database='utf8_general_ci'");

    return c;
}

void ts_connection_set_optimisation(ts_connection *c, int opt)
{
    c->opt_lvl = opt;
}

void ts_disconnect(ts_connection *c)
{
    if (c) {
	if (c->db) mysql_close(c->db);
	c->db = NULL;
	free(c->db_name);
	free(c);
    }
}

/* vi:set ts=8 sts=4 sw=4: */
