/*
 *  Copyright (C) 2002 Steve Harris, ECS, University of Southampton
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  $Id: import.c,v 1.33 2005/11/16 23:54:15 theno23 Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <string.h>
#include <mysql.h>
#include <mysqld_error.h>
#include <sys/types.h>
#include <sys/time.h>
#include <raptor.h>
#include <getopt.h>
#include <glib.h>

#include "config.h"
#include "tstore.h"
#include "inference.h"
#include "import.h"
#include "bulk-import.h"
#include "timing.h"
#include "rdf_constants.h"

#define MEMBER_PREFIX "http://www.w3.org/1999/02/22-rdf-syntax-ns#_"
#define MEMBER_PROP   "http://www.w3.org/2000/01/rdf-schema#member"

#define INTEGER_DT    "http://www.w3.org/2001/XMLSchema#integer"

#if MYSQL_VERSION_ID >= 40100
#define SQL_TRUE  "true"
#define SQL_FALSE "false"
#else
#define SQL_TRUE  "1"
#define SQL_FALSE "0"
#endif

#define TMP_SIZE 256

static void store_stmt(void *user_data, const raptor_statement *statement);

static void fatal_err_handler(void *data, raptor_locator *locator,
                          const char *message);
static void err_handler(void *data, raptor_locator *locator,
                          const char *message);

static void warning_handler(void *data, raptor_locator *locator,
                            const char *message);

/* recursivly remove data according to truth maintainance */
static int erase_data(ts_connection *db, ts_hash mhash, FILE *msg, int *tax_change, int verbosity);

static unsigned long long uid = 0; /* A unique(ish) id to append to genid's to
                                    * help avoid clashses */

#define CLASS_H_SIZE 1024       /* size of the running hash of classes that
				 * have been checked to see if thier declared */

typedef struct {
    int verbosity;
    int count_err;
    int count_warn;
    int count_trip;
    int tax_rebuild;
    ts_hash model;
    ts_connection *db;
    FILE *msg;
    ts_hash checked_classes[CLASS_H_SIZE];
} rs_parse_stuff;

void ts_orphan_cleanup(ts_connection *db)
{
    TIME(NULL);
    ts_mysql_query(db->db, "DELETE FROM "T_T" WHERE "
			"NOT EXISTS (SELECT * FROM "T_M" WHERE "
			T_T".tid="T_M".tid)");
    TIME("del_stale_triples");
}

int ts_erase(ts_connection *db, char *model, int flags, int verbosity, FILE *msg)
{
    char *query;
    ts_hash mhash = ts_ensure_resource(db, model);
    ts_hash sys_model = ts_ensure_resource(db, TS_SYS_MODEL);
    int ret = 0;

    if (flags & TS_MODEL_DATA) {
	int tax_change = 0;

	if (verbosity > 0) {
	    fprintf(msg, "removing model data\n");
	}
	ret = erase_data(db, mhash, msg, &tax_change, verbosity);
#ifndef ENABLE_QUADS
	if (!(flags & TS_MODEL_DELAY_CLEANUP)) {
	    ts_orphan_cleanup(db);
	}
#endif /* ! ENABLE_QUADS */
	if (tax_change) {
	    if (verbosity > 0) {
		fprintf(msg, "Rebuilding taxonomy\n");
	    }
	    ts_inference_setup(db, mhash);
	    ts_rebuild_taxonomy(db, verbosity);
	    ts_inference_commit(db);
	}
    }

    if (flags & TS_MODEL_HISTORY) {
	if (verbosity > 0) {
	    fprintf(msg, "removing model history\n");
	}
	query = g_strdup_printf("DELETE FROM "T_T" WHERE "
				"model=%lld AND subject=%lld",
				sys_model, mhash);
	if (mysql_query(db->db, query)) {
	    fprintf(msg, "Failed to remove history (%d): %s\n",
		    mysql_errno(db->db), mysql_error(db->db));
	    ret = 1;
	}
	free(query);
    }

    ts_bulk_cache_flush(db, verbosity, msg);

    return ret;
}

static int erase_data(ts_connection *db, ts_hash mhash, FILE *msg, int *tax_change, int verbosity)
{
    MYSQL_RES *res;
    MYSQL_ROW row;
    ts_hash sys_model = TS_ID(db, ts_hash_uri(TS_SYS_MODEL));
    ts_hash depends = TS_ID(db, ts_hash_uri(INT_DEPENDS));
    int ret = 0;
    char query[TMP_SIZE];

    mhash = TS_ID(db, mhash);

    if (verbosity > 1) {
	fprintf(msg, "   erasing from model %lld\n", mhash);
    }

    TIME(NULL);
    if (db->feature[TS_FEATURE_INFERENCE]) {
	ts_hash tran;

#ifndef ENABLE_QUADS
	snprintf(query, TMP_SIZE, "SELECT DISTINCT subject FROM "T_T", "
		T_M" WHERE "T_T".tid="T_M".tid "
		"AND mid=%lld AND predicate=%lld AND object=%lld",
		sys_model, depends, mhash);
#else /* ENABLE_QUADS */
	snprintf(query, TMP_SIZE, "SELECT DISTINCT subject FROM "
		T_T" WHERE model=%lld AND predicate=%lld AND "
		"object=%lld", sys_model, depends, mhash);
#endif /* ENABLE_QUADS */
	ts_mysql_query(db->db, query);
	res = mysql_store_result(db->db);
	while ((row = mysql_fetch_row(res))) {
	    ret += erase_data(db, atoll(row[0]), msg, tax_change, verbosity);
	}
	mysql_free_result(res);

#ifndef ENABLE_QUADS
	snprintf(query, TMP_SIZE, "DELETE FROM "T_T" WHERE "
		"tid=(SELECT tid FROM "T_M" WHERE mid=%lld) AND predicate=%lld "
		"AND object=%lld", sys_model, depends, mhash);
#else /* ENABLE_QUADS */
	snprintf(query, TMP_SIZE, "DELETE FROM "T_T" WHERE "
		"model=%lld AND predicate=%lld AND object=%lld", sys_model,
		depends, mhash);
#endif /* ENABLE_QUADS */
	ts_mysql_query(db->db, query);
	TIME("del_triples_recurse");

	/* check for transitive closure dependencies */
	tran = TS_ID(db, ts_hash_uri(INT_CLOSURE));
#ifndef ENABLE_QUADS
	snprintf(query, TMP_SIZE, "SELECT 1 FROM "T_T", "
	    T_M" WHERE "T_T".tid="T_M".tid "
	    "AND mid=%lld AND subject=%lld AND predicate=%lld AND "
	    "object=%lld LIMIT 1", sys_model, tran, depends, mhash);
#else
	snprintf(query, TMP_SIZE, "SELECT 1 FROM "T_T" "
	    "WHERE model=%lld AND subject=%lld AND predicate=%lld AND "
	    "object=%lld LIMIT 1", sys_model, tran, depends, mhash);
#endif
	if (ts_sql_count(db->db, query)) {
	    (*tax_change)++;
	    if (verbosity > 1) {
		fprintf(msg, "   removal requires TC rebuild\n");
	    }
	}
	TIME("tax_rebuild_check");
    }

#ifndef ENABLE_QUADS
    /* remove stale triples */
    snprintf(query, TMP_SIZE, "DELETE FROM "T_M" WHERE "
                            "mid=%lld", TS_ID(db, mhash));
#else /* ENABLE_QUADS */
    snprintf(query, TMP_SIZE, "DELETE FROM "T_T" WHERE "
			    "model=%lld", mhash);
#endif /* ENABLE_QUADS */
    ts_mysql_query(db->db, query);
    TIME("del_old_memberships");

    return ret;
}

int ts_erase_all(ts_connection *db, int verbosity, FILE *msg)
{
    ts_bulk_cache_flush(db, verbosity, msg);

    if (mysql_query(db->db, "DELETE FROM "T_T"")) {
	fprintf(msg, "Failed to remove triples (%d): %s\n",
		mysql_errno(db->db), mysql_error(db->db));
	return 1;
    }
    if (mysql_query(db->db, "DELETE FROM "T_S"")) {
	fprintf(msg, "Failed to remove symbols (%d): %s\n",
		mysql_errno(db->db), mysql_error(db->db));
	return 1;
    }
#ifndef ENABLE_QUADS
    if (mysql_query(db->db, "DELETE FROM "T_M"")) {
	fprintf(msg, "Failed to remove memberships (%d): %s\n",
		mysql_errno(db->db), mysql_error(db->db));
	return 1;
    }
#endif /* ! ENABLE_QUADS */
    ts_ensure_resource(db, NULL);
    ts_ensure_literal(db, NULL, 0, 0);
#ifndef ENABLE_QUADS
    ts_mysql_query(db->db, "ALTER TABLE "T_T" AUTO_INCREMENT=0;");
    ts_mysql_query(db->db, "ALTER TABLE "T_S" AUTO_INCREMENT=0;");
#endif /* ! ENABLE_QUADS */

    return 0;
}

int ts_import(ts_connection *db, char *uri, char *model, int verbosity,
    FILE *msg, ts_syntax format)
{
    char timestamp[TMP_SIZE];
    time_t t;
    struct tm *tp;
    char *poll_obj_uri;
    char *query;
    ts_hash sys_model;
    raptor_parser *rdf_parser = NULL;
    raptor_uri ruri, muri;
    rs_parse_stuff parse_data;

    parse_data.verbosity = verbosity;
    parse_data.db = db;
    parse_data.msg = msg;
    parse_data.tax_rebuild = 0;
    parse_data.count_err = 0;
    parse_data.count_warn = 0;
    parse_data.count_trip = 0;
    memset(parse_data.checked_classes, 0, CLASS_H_SIZE * sizeof(ts_hash));

    if (!uid) {
	struct timeval tv;

	gettimeofday(&tv, NULL);
	uid = (unsigned int) getpid();
	uid ^= ((unsigned long long) tv.tv_usec) << 16;
	uid ^= ((unsigned long long) tv.tv_sec) << 32;
    } else {
	uid++;
    }
    poll_obj_uri = g_strdup_printf("3store:po:%llx", uid);

    t = time(NULL);
    tp = localtime(&t);
    snprintf(timestamp, 255, "%04d-%02d-%02dT%02d:%02d:%02d", tp->tm_year +
	     1900, tp->tm_mon + 1, tp->tm_mday, tp->tm_hour, tp->tm_min,
	     tp->tm_sec);

    if (format == TS_AUTO) {
	if (strstr(uri, ".n3") || strstr(uri, ".ttl")) {
	    if (verbosity > 1) fprintf(msg, "using Turtle parser\n");
	    rdf_parser = raptor_new_parser("turtle");
	} else if (strstr(uri, ".nt")) {
	    if (verbosity > 1) fprintf(msg, "using NTriples parser\n");
	    rdf_parser = raptor_new_parser("ntriples");
	} else {
	    if (verbosity > 1) fprintf(msg, "using RDF/XML parser\n");
	    rdf_parser = raptor_new_parser("rdfxml");
	}
    } else if (format == TS_XML) {
	if (verbosity > 1) fprintf(msg, "using RDF/XML parser\n");
	rdf_parser = raptor_new_parser("rdfxml");
    } else if (format == TS_NTRIPLES) {
	if (verbosity > 1) fprintf(msg, "using NTriples parser\n");
	rdf_parser = raptor_new_parser("ntriples");
    } else if (format == TS_TURTLE) {
	if (verbosity > 1) fprintf(msg, "using Turtle parser\n");
	rdf_parser = raptor_new_parser("turtle");
    }
    if (!rdf_parser) {
	fprintf(msg, "Failed to create parser\n");
	return 1;
    }

    ts_assert_block_mode(db, TS_BLOCK_MODE_START);

    raptor_set_statement_handler(rdf_parser, &parse_data, store_stmt);
    raptor_set_warning_handler(rdf_parser, &parse_data, warning_handler);
    raptor_set_error_handler(rdf_parser, &parse_data, err_handler);
    raptor_set_fatal_error_handler(rdf_parser, &parse_data, fatal_err_handler);

    ruri = raptor_new_uri((unsigned char *)uri);

    if (!strcmp(model, "background")) {
	if (!db->feature[TS_FEATURE_STD_COMP]) {
	    fprintf(stderr, "Standards compliance mode is off, background graphs not supported\n");

	    return 1;
	}
	muri = raptor_new_uri((unsigned char *)"background:");
	parse_data.model = 0;
    } else {
	muri = raptor_new_uri((unsigned char *)model);
	parse_data.model = ts_ensure_resource(db, model);
    }

    ts_bulk_cache_flush(db, 0, msg);

    if (db->lock_tables) {
#ifdef ENABLE_QUADS
	ts_mysql_query(db->db, "LOCK TABLES "T_T" WRITE, "T_S" WRITE, "
		       T_I" WRITE, "T_F" WRITE, "T_A" WRITE, "
		       T_D" WRITE, "T_L" WRITE");
#else
	ts_mysql_query(db->db, "LOCK TABLES "T_T" WRITE, "T_S" WRITE, "
		       T_I" WRITE, "T_F" WRITE, "T_A" WRITE, "T_M" WRITE");
#endif
    }

    /* Process RDF input */
    parse_data.count_err = 0;
    parse_data.count_warn = 0;
    parse_data.count_trip = 0;
    if (raptor_parse_uri(rdf_parser, ruri, muri)) {
	fprintf(msg, "Failed to parse file\n");
	if (db->lock_tables && mysql_query(db->db, "UNLOCK TABLES")) {
	    fprintf(msg, "Failed to unlock tables (%d): %s\n",
		    mysql_errno(db->db), mysql_error(db->db));
	}

	return -1;
    }

    /* ensure that the pending table is complete */
    ts_assert_block_mode(db, TS_BLOCK_MODE_FLUSH);

    if (db->feature[TS_FEATURE_INFERENCE]) {
	ts_inference_setup(db, parse_data.model);
	ts_run_pending_inference(db);
	ts_inference_commit(db);
    }
   
    ts_assert_block_mode(db, TS_BLOCK_MODE_COMMIT);

    if (parse_data.tax_rebuild) {
	ts_inference_setup(db, parse_data.model);
	ts_rebuild_taxonomy(db, verbosity);
	ts_inference_commit(db);
    }

    /* Don't assert provenance for background graph loadings */
    if (parse_data.model && db->feature[TS_FEATURE_PROVENANCE]) {
	if (verbosity > 1) {
	    printf("Asserting provenance\n");
	}

	sys_model = ts_ensure_resource(db, TS_SYS_MODEL);
#ifndef ENABLE_QUADS
	query = g_strdup_printf("DELETE FROM "T_T" WHERE "
				"tid=(SELECT tid FROM "T_M" WHERE mid=%lld) "
				"AND subject=%lld AND predicate=%lld",
			TS_ID(db, sys_model), TS_ID(db, db->model),
			TS_ID(db, ts_hash_uri(INT_NS "last-polled")));
#else /* ENABLE_QUADS */
	query = g_strdup_printf("DELETE FROM "T_T" WHERE "
				"model=%lld AND subject=%lld AND "
				"predicate=%lld",
			sys_model, db->model,
			ts_hash_uri(INT_NS "last-polled"));
#endif /* ENABLE_QUADS */
	if (mysql_query(db->db, query)) {
	    fprintf(msg, "Warning: failed to remove old timestamp '%s' (%d): "
		    "%s\n", query, mysql_errno(db->db), mysql_error(db->db));
	}
	free(query);

	ts_assert_triple(db, sys_model, model,
		         "http://purl.org/dc/elements/1.1/source", uri, ObjURI,
			 0, 0);
	ts_assert_triple(db, sys_model, model, INT_NS "last-polled",
		         poll_obj_uri, ObjURI, 0, 0);
	ts_assert_triple(db, sys_model, model, INT_NS "polled", poll_obj_uri,
		         ObjURI, 0, 0);
	ts_assert_triple(db, sys_model, poll_obj_uri, INT_NS "date", timestamp,
			 ObjLiteral, 0, 0);
	if (parse_data.count_err == 0 && parse_data.count_warn == 0) {
	    ts_assert_triple(db, sys_model, poll_obj_uri,
		             INT_NS "status-message", "OK", ObjLiteral, 0, 0);
	    ts_assert_triple(db, sys_model, poll_obj_uri, INT_NS "status",
			     INT_NS "success", ObjURI, 0, 0);
	} else {
	    char msg[TMP_SIZE];

	    snprintf(msg, 255, "%d error%s and %d warning%s",
		     parse_data.count_err, parse_data.count_err == 1 ? "" : "s",
		     parse_data.count_warn, parse_data.count_warn == 1 ?
		                            "" : "s");
	    ts_assert_triple(db, sys_model, poll_obj_uri,
		             INT_NS "status-message", msg, ObjLiteral, 0, 0);
	    if (parse_data.count_err > 0) {
		ts_assert_triple(db, sys_model, poll_obj_uri, INT_NS "status",
				 INT_NS "failure", ObjURI, 0, 0);
	    } else {
		ts_assert_triple(db, sys_model, poll_obj_uri, INT_NS "status",
				 INT_NS "warning", ObjURI, 0, 0);
	    }
	}
    }

    ts_mysql_query(db->db, "UNLOCK TABLES");

    if (parse_data.count_err) {
	return -parse_data.count_err;
    } else {
	return parse_data.count_trip;
    }
}

static void store_stmt(void *user_data, const raptor_statement *statement)
{
    rs_parse_stuff *data = (rs_parse_stuff *)user_data;
    ts_obj_type type;
    char tmps[TMP_SIZE], tmpp[TMP_SIZE], tmpo[TMP_SIZE];
    char *subj = (char *)raptor_uri_as_string((raptor_uri *)statement->subject),
	 *pred = (char *)raptor_uri_as_string((raptor_uri *)statement->predicate),
         *obj;
    const char *datatype = (char *)
	raptor_uri_as_string(statement->object_literal_datatype);
    int dt = 0;
    int lang = 0;
    /* pick a bnode suffix, if we have a model ID, use that, otherwise (ie.
     * were in the background graph, usea unique-ish int */
    long long bnodesuf = data->model ? data->model : uid;

    TIME("unknown");

    if (datatype) {
#ifdef ENABLE_QUADS
	dt = ts_ensure_datatype(data->db, (char *)datatype);
#else
	dt = ts_ensure_resource(data->db, datatype);
#endif
    }
    if (statement->object_literal_language) {
        char *langtag = (char *)statement->object_literal_language;

#ifdef ENABLE_QUADS
	lang = ts_ensure_language(data->db, langtag);
#else
	lang = ts_ensure_literal(data->db, langtag, 0, 0);
#endif
    }

    (data->count_trip)++;
    if (statement->object_type == RAPTOR_IDENTIFIER_TYPE_LITERAL ||
	statement->object_type == RAPTOR_IDENTIFIER_TYPE_XML_LITERAL) {
	type = ObjLiteral;
	obj = (char *)statement->object;
    } else {
	type = ObjURI;
	obj = (char *)raptor_uri_as_string((raptor_uri *)statement->object);
    }

    if (statement->subject_type == RAPTOR_IDENTIFIER_TYPE_ANONYMOUS) {
	sprintf(tmps, "_:%s.%llx", (char *) statement->subject, bnodesuf);
	subj = tmps;
    }
    if (statement->predicate_type == RAPTOR_IDENTIFIER_TYPE_ANONYMOUS) {
	sprintf(tmpp, "_:%s.%llx", (char *) statement->predicate, bnodesuf);
	pred = tmpp;
    }
    if (statement->object_type == RAPTOR_IDENTIFIER_TYPE_ANONYMOUS) {
	sprintf(tmpo, "_:%s.%llx", (char *) statement->object, bnodesuf);
	obj = tmpo;
    }

    if (statement->predicate_type == RAPTOR_IDENTIFIER_TYPE_ORDINAL) {
	sprintf(tmpp, MEMBER_PREFIX "%d", *((int *)pred));
	pred = tmpp;
    }

    if (data->verbosity > 1) {
	fprintf(stdout, "asserting (<%s>, <%s>, %c%s%c), dt:%s, lang:%s\n",
                subj, pred,
	        type == ObjLiteral ? '"' : '<', obj,
	        type == ObjLiteral ? '"' : '>', (char *)datatype,
		(char *)statement->object_literal_language);
    }

    TIME("parser_handler");
    if (ts_assert_triple(data->db, data->model, subj, pred, obj, type, dt,
	lang) == TS_BAD_SQL) {
	fprintf(data->msg, "tstore_import: Bad SQL, skipping statement\n");
    }
    TIME("assert");

    if (statement->predicate_type == RAPTOR_IDENTIFIER_TYPE_ORDINAL) {
	if (ts_assert_triple(data->db, data->model, subj, MEMBER_PROP, obj,
	    type, dt, lang) == TS_BAD_SQL) {
	    fprintf(data->msg, "tstore_import: Bad SQL, skipping entailment\n");
	}
	if (ts_assert_triple(data->db, data->model, pred, RDFS_SUBPROPERTY,
			     MEMBER_PROP, ObjURI, 0, 0) == TS_BAD_SQL) {
	    fprintf(data->msg, "tstore_import: Bad SQL, skipping entailment\n");
	}
    }

    /* checks for pending inference */
    if (data->db->feature[TS_FEATURE_INFERENCE] && !data->tax_rebuild) {
	if (!strcmp(pred, RDFS_SUBCLASS) || !strcmp(pred, RDFS_SUBPROPERTY)) {
	    data->tax_rebuild = 1;
	} else if (!strcmp(pred, RDF_TYPE)) {
	    ts_hash class = ts_hash_uri(obj);

	    /* if we may not have seen this class */
	    if (data->checked_classes[class & (CLASS_H_SIZE - 1)] != class) {
		char tmp[TMP_SIZE];
		int count;

		data->checked_classes[class & (CLASS_H_SIZE - 1)] = class;
		snprintf(tmp, TMP_SIZE, "SELECT 1 FROM "T_T" WHERE "
			 "subject=%lld AND predicate=-5275523445234180533 AND "
			 "object=%lld LIMIT 1", class, class);
		count = ts_sql_count(data->db->db, tmp);
		if (count == 0) {
		    ts_hash tc_hash = TS_ID(data->db, ts_ensure_resource
			    (data->db, INT_CLOSURE));
		    /* shortcut the taxonomic closure code incase theres no
		     * class graph. If the taxonomic closure is recalculated
		     * later this will be overwritten anyway */
		    ts_assert_triple(data->db, tc_hash, obj, RDFS_SUBCLASS,
				     obj, ObjURI, 0, 2);
		    ts_assert_triple(data->db, tc_hash, obj, RDF_TYPE,
				     RDFS_CLASS, ObjURI, 0, 2);
		}
	    }
	}
    }
    TIME("parser_handler");
}

static void
fatal_err_handler(void *user_data, raptor_locator * locator,
                    const char *message)
{
    rs_parse_stuff *data = (rs_parse_stuff *)user_data;

    fprintf(data->msg, "3store import fatal error: ");
    raptor_print_locator(data->msg, locator);
    fprintf(data->msg, " - %s\n", message);
    (data->count_err)++;
}

static void
err_handler(void *user_data, raptor_locator * locator, const char *message)
{
    rs_parse_stuff *data = (rs_parse_stuff *)user_data;

    fprintf(data->msg, "3store import error: ");
    raptor_print_locator(data->msg, locator);
    fprintf(data->msg, " - %s\n", message);
    (data->count_err)++;
}

static void
warning_handler(void *user_data, raptor_locator * locator, const char *message)
{
    rs_parse_stuff *data = (rs_parse_stuff *)user_data;

    fprintf(data->msg, "3store import warning: ");
    raptor_print_locator(stderr, locator);
    fprintf(data->msg, " - %s\n", message);
    (data->count_warn)++;
}

/* vi:set ts=8 sts=4 sw=4: */
