/*
 *  Copyright (C) 2005 Steve Harris
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  $Id: inference.c,v 1.11 2005/11/17 08:10:28 theno23 Exp $
 */

#include <mysql.h>

#include "tstore.h"
#include "import.h"
#include "util.h"
#include "timing.h"
#include "config.h"

#define BLOCK_SIZE 1024

#define TMP_SIZE 256

#define rs_mat_hash(m, a, b) ((a * (m->size + 1) + b) % m->hash_size)

#define ts_cr_model_key(r, c) ((r * 65536) + c)

/* Sparse matrix representation structures */

/* List of all the [a,b]'s for which the value is true */
typedef struct _fl {
    int a;
    int b;
    struct _fl *next;
    struct _fl *next_hash;
} rs_full_list;

/* List of all the b's for this a for which the value is true */
typedef struct _sl {
    int b;
    struct _sl *next;
} rs_split_list;

/* Overall struct representing the matrix */
typedef struct {
    rs_full_list *l_current;
    rs_full_list *l_pending;
    rs_split_list **s_current;
    rs_split_list **s_pending;
    rs_full_list **hash;
    int size;
    int hash_size;
} rs_sparse_mat;

typedef struct rs_raw_triple_s {
        ts_hash modl;
        ts_hash subj;
        ts_hash pred;
        ts_hash obj;
} rs_raw_triple;

/* Matrix handling functions */

/*create new matrix */
rs_sparse_mat *rs_mat_new(int size);

/* set [a,b] to true */
int rs_mat_set(rs_sparse_mat *m, int a, int b);

gint hash_cmp(gconstpointer a, gconstpointer b);

gboolean rs_hash_equal(gconstpointer a, gconstpointer b);
guint rs_hash_hash(gconstpointer v);

static void rs_block_append(ts_connection *db, rs_raw_triple *block, int *count,
                             ts_hash m, ts_hash s, ts_hash p, ts_hash o);
static void rs_block_flush(ts_connection *db, rs_raw_triple *block, int *count);
static int rs_assert_raw_block(ts_connection *this, rs_raw_triple *block,
                        int count);

/* create a virtual mode to hold triples that are infered from model a and b */
static ts_hash ts_ensure_virtual_model_2(ts_connection *c, ts_hash a, ts_hash b);

/* rebuild_latice_inline:
 *
 * latice rebuilting code. This produces flagged triples in the main triples
 * table.
 *
 * type: the rdf:type of the individulas that are the nodes on the latice.
 * rel: the relation that makes up the edges of the latice
 * top: the latice top, or NULL if it is free
 * bottom: the lattice bottom, or NULL if it is free
 */
static void rebuild_latice_inline(ts_connection * db, char *type, char *rel,
				  char *top, char *bottom, int verbosity);

void ts_inference_setup(ts_connection *c, ts_hash model)
{
    char tmp[TMP_SIZE];

    ts_mysql_query(c->db, "CREATE TEMPORARY TABLE infered (ma bigint NOT NULL, mb bigint NOT NULL, s bigint NOT NULL, p bigint NOT NULL, o bigint NOT NULL)");
#ifndef ENABLE_QUADS
    ts_mysql_query(c->db, "CREATE TEMPORARY TABLE pending_i "
			"(s int NOT NULL, p int NOT NULL, o int NOT NULL, "
			"m int NOT NULL, t INT NOT NULL)");
#endif /* ! ENABLE_QUADS */
    
    snprintf(tmp, TMP_SIZE, "SET @pmodel=%lld", model);
    ts_mysql_query(c->db, tmp);
}

/* produces inferred statements from between the pending table and the existing
 * triples only */
void ts_run_pending_inference(ts_connection *c, ts_hash model)
{
    /* [cough] I'm sorry about this */
    #include "simple-rules.c"
}

void ts_inference_commit(ts_connection *c)
{
    MYSQL_RES *res;
    MYSQL_ROW row;
    char tmp[512];

    /* Determine what virtual graph inferred triples belong in and create the
     * graph where neccesary */
    ts_mysql_query(c->db, "SELECT DISTINCT ma, mb FROM infered WHERE mb != 1 && ma != mb");
    res = mysql_store_result(c->db);
    while ((row = mysql_fetch_row(res))) {
	ts_hash ha = atoll(row[0]);
	ts_hash hb = atoll(row[1]);
	ts_hash hv; /* hash of the virtual model */

	hv = ts_ensure_virtual_model_2(c, ha, hb);

	snprintf(tmp, 511, "UPDATE infered SET ma=0x%llx, mb=1 "
		 "WHERE ma=0x%llx AND mb=0x%llx", hv, ha, hb);
	ts_mysql_query(c->db, tmp);
    }

#ifndef ENABLE_QUADS
    ts_mysql_query(c->db, "INSERT INTO pending_i "
		   "SELECT DISTINCT s, p, o, ma, 0 FROM infered");
#else /* ENABLE_QUADS */
    ts_mysql_query(c->db, "INSERT INTO "T_T" SELECT DISTINCT ma, s, p, o "
		          "FROM infered");
#endif /* ENABLE_QUADS */
    ts_mysql_query(c->db, "DROP TABLE infered");

#ifndef ENABLE_QUADS
    //XXX potential optimisation: it may be possible to do this from the
    //    inferred table, instead of going via another pending table
    ts_assert_block_mode(c, TS_BLOCK_MODE_COMMIT_INF);
#endif /* ! ENABLE_QUADS */
}

void ts_rebuild_taxonomy(ts_connection *db, int verbosity)
{
    ts_erase(db, INT_CLOSURE, TS_MODEL_DATA, verbosity, stderr);

    TIME(NULL);
    if (verbosity > 0) printf("finding entailments for rdfs5\n");
    /* Cacluate transitive closure of rdfs:subPropertyOf relation */
    rebuild_latice_inline(db, RDF_PROPERTY, RDFS_SUBPROPERTY, NULL, NULL,
			  verbosity);

    if (verbosity > 0) printf("finding entailments for rdfs7+8\n");
    /* Cacluate transitive closure of rdf:subClassOf relation */
    rebuild_latice_inline(db, RDFS_CLASS, RDFS_SUBCLASS, RDFS_RESOURCE, NULL,
			  verbosity);
    TIME("latice_rebuild");
}

static void model_mark(gpointer key, gpointer value, gpointer user_data)
{
    rs_raw_triple *raw_block = ((rs_raw_triple **)user_data)[0];
    int *count = ((int **)user_data)[1];
    ts_connection *c = ((ts_connection **)user_data)[2];

    ts_hash depends = TS_ID(c, ts_hash_uri(INT_DEPENDS));
    ts_hash tran = TS_ID(c, ts_hash_uri(INT_CLOSURE));
    ts_hash sys_model = TS_ID(c, ts_hash_uri(TS_SYS_MODEL));

    rs_block_append(c, raw_block, count, sys_model, tran, depends, *((int64_t *)value));
}

static void rebuild_latice_inline(ts_connection *db, char *type, char *rel,
				  char *top, char *bottom, int verbosity)
{
    MYSQL_RES *mres;
    MYSQL_ROW mrow;
    /* A list of the individuals in the latice */
    GSList *indivs = NULL;
    GSList *tmp = NULL, *it;
    GHashTable *rc_model_ht = g_hash_table_new(rs_hash_hash, rs_hash_equal);
    int num_indivs;
    unsigned long i;
    rs_sparse_mat *mat;
    GHashTable *uri_lookup;
    ts_hash *skey, *okey;
    ts_hash top_key, bottom_key;
    ts_hash *rc_hash;
    ts_hash rel_hash = TS_ID(db, ts_hash_uri(rel));
    ts_hash type_hash = TS_ID(db, ts_hash_uri(type));
    ts_hash rdf_type_hash = TS_ID(db, ts_hash_uri(RDF_TYPE));
    ts_hash tc_model = TS_ID(db, ts_hash_uri(INT_CLOSURE));
    rs_raw_triple raw_block[BLOCK_SIZE];
    int block_count = 0;
    int count;
    rs_full_list *it_f;
    rs_split_list *it_s;
    char query[TMP_SIZE];

    if (top) ts_ensure_resource(db, top);
    if (bottom) ts_ensure_resource(db, bottom);

    /* Find the individuals that are expliticly of the right type */
    snprintf(query, TMP_SIZE-1, "SELECT subject FROM "T_T" WHERE predicate=%lld AND object=%lld", rdf_type_hash, type_hash);
    ts_mysql_query(db->db, query);
    mres = mysql_store_result(db->db);
    while ((mrow = mysql_fetch_row(mres))) {
	skey = malloc(sizeof(ts_hash));
	*skey = atoll(mrow[0]);
	indivs = g_slist_prepend(indivs, skey);
    }
    mysql_free_result(mres);

    /* Find the individuals that are implied to be of the right type
     * becuase of the relation + rdfs[23] */
    snprintf(query, TMP_SIZE-1, "SELECT subject, object FROM "T_T" WHERE predicate=%lld", rel_hash);
    ts_mysql_query(db->db, query);
    mres = mysql_store_result(db->db);
    while ((mrow = mysql_fetch_row(mres))) {
	skey = malloc(sizeof(ts_hash));
	*skey = atoll(mrow[0]);
	okey = malloc(sizeof(ts_hash));
	*okey = atoll(mrow[1]);

	indivs = g_slist_prepend(indivs, skey);
	indivs = g_slist_prepend(indivs, okey);
    }
    mysql_free_result(mres);

    /* sort+uniq the list to remove duplicates */
    indivs = g_slist_sort(indivs, hash_cmp);

    for (it = indivs; it; it = it->next) {
	if (!it->next || *(ts_hash *)(it->data) != *(ts_hash *)(it->next->data)) {
	    tmp = g_slist_prepend(tmp, it->data);
	}
    }
    g_slist_free(indivs);
    indivs = tmp;
    num_indivs = g_slist_length(indivs);

    rc_hash = malloc(num_indivs * sizeof(ts_hash));
    uri_lookup = g_hash_table_new(rs_hash_hash, rs_hash_equal);
    for (i = 0, it = indivs; it; it = it->next, i++) {
	g_hash_table_insert(uri_lookup, it->data, (gpointer) i);
	rc_hash[i] = *(ts_hash *)(it->data);
	//printf("%s\n", (char *)it->data);
    }

    top_key = TS_ID(db, ts_hash_uri(top));
    bottom_key = TS_ID(db, ts_hash_uri(bottom));

    for (it = indivs; it; it=it->next) {
	ts_hash indiv = *(ts_hash *)(it->data);

	/* Assert that a foo is a subfoo of itsself */
	rs_block_append(db, raw_block, &block_count, tc_model,
			indiv, rel_hash, indiv);

	/* Assert that a foo is of whatever type */
	rs_block_append(db, raw_block, &block_count, tc_model,
			indiv, rdf_type_hash, type_hash);

	if (top) {
	    /* Assert that a foo is a subfoo of top */
	    rs_block_append(db, raw_block, &block_count, tc_model,
			indiv, rel_hash, top_key);
	}
	if (bottom) {
	    /* Assert that a foo is a superfoo of bottom */
	    rs_block_append(db, raw_block, &block_count, tc_model,
			bottom_key, rel_hash, indiv);
	}
    }
    rs_block_flush(db, raw_block, &block_count);

    mat = rs_mat_new(num_indivs);

    ts_mysql_query(db->db, "UNLOCK TABLES");

    /* Find ground facts about relation and populate the matrix with them */
#ifndef ENABLE_QUADS
    snprintf(query, TMP_SIZE-1, "SELECT DISTINCT subject, object, mid FROM "T_T", "T_M" WHERE "T_T".tid="T_M".tid AND predicate=%lld AND mid!=%lld", rel_hash, tc_model);
#else /* ENABLE_QUADS */
    snprintf(query, TMP_SIZE-1, "SELECT DISTINCT subject, object, model FROM "T_T" WHERE predicate=%lld AND model!=%lld", rel_hash, tc_model);
#endif /* ENABLE_QUADS */
    ts_mysql_query(db->db, query);
    mres = mysql_store_result(db->db);
    while ((mrow = mysql_fetch_row(mres))) {
	long c, r;
	int32_t  hash;
	ts_hash h;
	ts_hash *m;

	/* The sub-foo is the matrix column */
	h = atoll(mrow[0]);
	c = (long) g_hash_table_lookup(uri_lookup, &h);
	/* And the super-foo is the row */
	h = atoll(mrow[1]);
	r = (long) g_hash_table_lookup(uri_lookup, &h);

	rs_mat_set(mat, c, r);
	m = malloc(sizeof(ts_hash));
	*m = atoll(mrow[2]);
	hash = ts_cr_model_key(c, r);
	g_hash_table_insert(rc_model_ht, &hash, m);
    }

    /* Floyd-Warshall's algorithm:
     *
     * adjacent[X][Z] and adjacent[Z][Y] => adjacent[X][Y]
     */

    while (mat->l_pending) {
	ts_hash *m1, *m2, *m;
	int hash;
        /* roll pending and current matricies */
        mat->l_current = mat->l_pending;
        mat->l_pending = NULL;

        for (i=0; i<mat->size; i++) {
            mat->s_current[i] = mat->s_pending[i];
            mat->s_pending[i] = NULL;
        }

        count = 0;
        for (it_f = mat->l_current; it_f; it_f = it_f->next) {
            for (it_s = mat->s_current[it_f->b]; it_s; it_s = it_s->next) {
                if (rs_mat_set(mat, it_f->a, it_s->b)) {
                    /* assert */
		    hash = ts_cr_model_key(it_f->a, it_f->b);
		    m1 = g_hash_table_lookup(rc_model_ht, &hash);
		    hash = ts_cr_model_key(it_f->b, it_s->b);
		    m2 = g_hash_table_lookup(rc_model_ht, &hash);
		    m = malloc(sizeof(ts_hash));
		    if (m1 && !m2) {
			*m = TS_ID(db, ts_ensure_virtual_model_2(db, *m1, *m1));
		    } else if (m1 && m2) {
			*m = TS_ID(db, ts_ensure_virtual_model_2(db, *m1, *m2));
		    }
		    hash = ts_cr_model_key(it_f->a, it_s->b);
		    g_hash_table_insert(rc_model_ht, &hash, m);
		    if (verbosity > 1) {
			printf("A1 %d,%d %d,%d\n", it_f->a, it_f->b, it_f->b, it_s->b);
		    }
		    rs_block_append(db, raw_block, &block_count, *m,
				    rc_hash[it_f->a], rel_hash,
                                    rc_hash[it_s->b]);
                    count++;
                }
            }
        }

        for (it_f = mat->l_current; it_f; it_f = it_f->next) {
            for (it_s = mat->s_pending[it_f->b]; it_s; it_s = it_s->next) {
                if (rs_mat_set(mat, it_f->a, it_s->b)) {
                    /* assert */
		    hash = ts_cr_model_key(it_f->a, it_f->b);
		    m1 = g_hash_table_lookup(rc_model_ht, &hash);
		    hash = ts_cr_model_key(it_f->b, it_s->b);
		    m2 = g_hash_table_lookup(rc_model_ht, &hash);
		    m = malloc(sizeof(ts_hash));
		    *m = TS_ID(db, ts_ensure_virtual_model_2(db, *m1, *m2));
		    hash = ts_cr_model_key(it_f->a, it_s->b);
		    g_hash_table_insert(rc_model_ht, &hash, m);
		    if (verbosity > 1) {
			printf("A2 %d,%d %d,%d\n", it_f->a, it_f->b, it_f->b, it_s->b);
		    }
		    rs_block_append(db, raw_block, &block_count, *m,
				    rc_hash[it_f->a], rel_hash,
                                    rc_hash[it_s->b]);
                    count++;
                }
            }
        }
	if (verbosity > 1) {
	    printf("added %d\n", count);
	}
    }

    ts_ensure_resource(db, INT_DEPENDS);
    ts_ensure_resource(db, INT_CLOSURE);

    {
	void *ptrs[3] = {raw_block, &block_count, db};
	g_hash_table_foreach(rc_model_ht, model_mark, ptrs);
    }

    rs_block_flush(db, raw_block, &block_count);
}

gint hash_cmp(gconstpointer a, gconstpointer b)
{
    ts_hash av, bv;

    av = *(ts_hash *)a;
    bv = *(ts_hash *)b;

    if (av == bv) {
	return 0;
    } else if (av > bv) {
	return 1;
    }

    return -1;
}

rs_sparse_mat *rs_mat_new(int size)
{
    rs_sparse_mat *m;

    m = malloc(sizeof(rs_sparse_mat));
    m->size = size;
    m->hash_size = size * 5;
    m->l_current = NULL;
    m->l_pending = NULL;
    m->s_current = calloc(size, sizeof(rs_split_list));
    m->s_pending = calloc(size, sizeof(rs_split_list));
    m->hash = calloc(m->hash_size, sizeof(rs_full_list));

    return m;
}

int rs_mat_set(rs_sparse_mat *m, int a, int b)
{
    rs_split_list *it;
    rs_full_list *ni, *hash_pt;

    hash_pt = m->hash[rs_mat_hash(m, a, b)];
    for (ni = hash_pt; ni; ni=ni->next_hash) {
        if (ni->a == a && ni->b == b) {
            /* Its allready set */
            return 0;
        }
    }

    /* Its not been set allready, set it */
    it = malloc(sizeof(rs_split_list));
    it->b = b;
    it->next = m->s_pending[a];
    m->s_pending[a] = it;

    ni = malloc(sizeof(rs_full_list));
    ni->a = a;
    ni->b = b;
    ni->next = m->l_pending;
    m->l_pending = ni;

    /* The hash shares data with the list (to save bytes), so we need to
     * set its 'next' ptr */
    ni->next_hash = hash_pt;
    m->hash[rs_mat_hash(m, a, b)] = ni;

    return 1;
}

gboolean rs_hash_equal(gconstpointer a, gconstpointer b)
{
    return (*(ts_hash *)a == *(ts_hash *)b);
}

guint rs_hash_hash(gconstpointer v)
{
    return *(ts_hash *)v;
}

static void rs_block_append(ts_connection *db, rs_raw_triple *block, int *count,
                            ts_hash m, ts_hash s, ts_hash p, ts_hash o)
{
    block[*count].modl = m;
    block[*count].subj = s;
    block[*count].pred = p;
    block[*count].obj = o;
    if (++*count == BLOCK_SIZE) {
	rs_block_flush(db, block, count);
    }
}

static void rs_block_flush(ts_connection *db, rs_raw_triple *block, int *count)
{
    if (*count > 0) {
	if (rs_assert_raw_block(db, block, *count)) {
	    fprintf(stderr, "Fatal error, exiting\n");
	    exit(1);
	}
    }
    *count = 0;
}

int rs_assert_raw_block(ts_connection *this, rs_raw_triple *block,
                        int count)
{
    char *query;
    int ofs;
    int length;
    int i;

    length = 100 + 110 * count;
    query = malloc(length * sizeof(char));
#ifndef ENABLE_QUADS
    ofs = sprintf(query, "INSERT INTO pending_i VALUES ");
#else /* ENABLE_QUADS */
    ofs = sprintf(query, "INSERT INTO "T_T" VALUES ");
#endif /* ENABLE_QUADS */
    for (i = 0; i < count; i++) {
#ifndef ENABLE_QUADS
        ofs += snprintf(query + ofs, 110, "(%lld,%lld,%lld,%lld,0)%s",
                     block[i].subj, block[i].pred, block[i].obj, block[i].modl,
		     i == count - 1 ? "" : ",");
#else /* ENABLE_QUADS */
        ofs += snprintf(query + ofs, 110, "(%lld,%lld,%lld,%lld)%s",
                     block[i].modl, block[i].subj, block[i].pred, block[i].obj,
		     i == count - 1 ? "" : ",");
#endif /* ENABLE_QUADS */
    }
    if (ofs >= length) {
        fprintf(stderr, "Buffer overrun in %s:%d\n", __FILE__, __LINE__);
        return 1;
    }

    ts_mysql_query(this->db, query);
    free(query);

    return 0;
}

static ts_hash ts_ensure_virtual_model_2(ts_connection *c, ts_hash a, ts_hash b)
{
    ts_hash hv;
    ts_hash sys_model = ts_ensure_resource(c, TS_SYS_MODEL);
    ts_hash depends = ts_ensure_resource(c, INT_DEPENDS);
    char tmp[TMP_SIZE];

    if (a == b) {
	/* really its a single dependency */
	snprintf(tmp, TMP_SIZE-1, "_:vg%llx", a);
	hv = ts_ensure_resource(c, tmp);

	return hv;
    }

    /* create a deterministic ID for the virtual model */
    if (a < b) {
	snprintf(tmp, TMP_SIZE-1, "_:vg%llx.%llx", a, b);
    } else {
	snprintf(tmp, TMP_SIZE-1, "_:vg%llx.%llx", b, a);
    }
    hv = ts_ensure_resource(c, tmp);

    ts_assert_raw(c, TS_ID(c, sys_model), TS_ID(c, hv), TS_ID(c, depends), a);
    ts_assert_raw(c, TS_ID(c, sys_model), TS_ID(c, hv), TS_ID(c, depends), b);

    return hv;
}

/* vi:set ts=8 sts=4 sw=4: */
