/*
 *  Copyright (C) 2005 Steve Harris
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  $Id: query.c,v 1.13 2005/12/06 13:57:58 theno23 Exp $
 */

#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include <math.h>
#include <mysql.h>
#include <rasqal.h>

#include "tstore.h"

void parser_error_handler(void *ud, raptor_locator* locator, const char *msg)
{
    ts_query *tq = ud;
    char locstr[1024];
    char *newstr;

    locstr[0] = '\0';
    raptor_format_locator(locstr, sizeof(locstr), locator);
    newstr = g_strdup_printf("%serror at %s - %s\n",
			     tq->error_str ? tq->error_str : "",
			     locstr, msg);
    if (tq->error_str) {
	free(tq->error_str);
    }
    tq->error_str = newstr;
}

ts_query *ts_query_prepare(ts_connection *c, const char *lang,
	const char *query, const char *base_uri)
{
    rasqal_query *rq;
    raptor_uri *bu;
    ts_query *tq;
    int ret;
    raptor_sequence *vars;
    raptor_sequence *froms;
    unsigned int i;
    int colno = 0;

    tq = calloc(1, sizeof(ts_query));
    bu = raptor_new_uri((unsigned char *)base_uri);
    rq = rasqal_new_query(lang, NULL);
    rasqal_query_set_error_handler(rq, tq, parser_error_handler);
    ret = rasqal_query_prepare(rq, (unsigned char *)query, bu);
    if (ret) {
	return tq;
    }

    froms = rasqal_query_get_data_graph_sequence(rq);
    for (i=0; 1; i++) {
	rasqal_data_graph *dg = raptor_sequence_get_at(froms, i);
	char *guri;
	int exists;
	char *gquery;

	if (!dg) {
	    break;
	}

	if (c->feature[TS_FEATURE_STD_COMP]) {
	    if (dg->flags == RASQAL_DATA_GRAPH_BACKGROUND) {
		ts_import(c, (char *)raptor_uri_as_string(dg->uri),
		          "background:", 1, stderr, TS_AUTO);
		tq->glimit = g_slist_prepend(tq->glimit,
						raptor_uri_to_string(dg->uri));
		continue;
	    }
	}

	guri = dg->name_uri ? (char *)raptor_uri_to_string(dg->name_uri) :
			      (char *)raptor_uri_to_string(dg->uri);

	/* look for data allready in graph */
#ifdef ENABLE_QUADS
	gquery = g_strdup_printf("SELECT 1 FROM "T_T" WHERE model=%lld "
				 "LIMIT 1", TS_ID(c, ts_hash_uri(guri)));
#else
	gquery = g_strdup_printf("SELECT 1 FROM "T_M" WHERE mid=%lld LIMIT 1",
				 TS_ID(c, ts_hash_uri(guri)));
#endif
	exists = ts_sql_count(c->db, gquery);
	free(gquery);
	if (!exists) {
	    if (raptor_sequence_size(froms) > 0 &&
		!(c->feature[TS_FEATURE_LOADING])) {
		tq->error_str = strdup("FROM or FROM NAMED used to load from "
			"URI, but that feature is disabled for security "
			"reasons. If you wish to eable it reconfigure with "
			"ts-setup.");

		return tq;
	    }
	    ts_import(c, (char *)raptor_uri_as_string(dg->uri), guri, 1,
		      stderr, TS_AUTO);
	}

	tq->glimit = g_slist_prepend(tq->glimit, guri);
    }

    tq = ts_sql_compile(tq, c, rq);
    if (tq->error_str) {
	return tq;
    }
    tq->rquery = rq;
    tq->base = bu;

    vars = tq->vars;
    tq->cols = vars ? raptor_sequence_size(vars): 0;
    tq->col_desc = calloc(tq->cols, sizeof(ts_binding_desc));
    tq->bindings = calloc(tq->cols, sizeof(ts_binding));
    for (i = 0; i < tq->cols; i++) {
	rasqal_variable *v = raptor_sequence_get_at(vars, i);
	ts_symbol *s = ts_symtab_get(tq->symtab, (unsigned char *)v->name, 0, 0);

	tq->col_desc[i].sym = (char *)v->name;
	tq->col_desc[i].col = colno++;
	if (!s->bound_in_msp || c->opt_lvl < 2) {
	    tq->col_desc[i].dt_col = colno++;
	    tq->col_desc[i].lang_col = colno++;
	} else {
	    tq->col_desc[i].dt_col = -1;
	    tq->col_desc[i].lang_col = -1;
	}
    }

    tq->distinct = rasqal_query_get_distinct(rq);
    ts_symtab_free(tq->symtab);
    tq->symtab = NULL;

    return tq;
}

ts_result *ts_query_execute(ts_connection *c, ts_query *q)
{
    GSList *it;
    int i;
    ts_result *r = calloc(1, sizeof(ts_result));

    if (!c) {
	return NULL;
    }
    r->con = c;
    if (q->error_str) {
	r->error_str = g_strdup(q->error_str);

	return r;
    }

    r->query = q;

    /* execute pre query stage */
    for (it = q->sql_setup; it; it = it->next) {
        if (mysql_query(c->db, it->data)) {
	    r->error_str = g_strdup_printf("SQL error in '%s' %d: %s\n",
		    (char *)it->data, mysql_errno(c->db), mysql_error(c->db));

	    return r;
	}
    }
    for (it = q->sql_pre_query; it; it = it->next) {
        if (mysql_query(c->db, it->data)) {
	    r->error_str = g_strdup_printf("SQL error in '%s' %d: %s\n",
		    (char *)it->data, mysql_errno(c->db), mysql_error(c->db));

	    return r;
	}
    }

    /* execute actual query */
    if (mysql_query(c->db, q->sql_query)) {
	r->error_str = g_strdup_printf("SQL error in '%s' %d: %s\n",
		q->sql_query, mysql_errno(c->db), mysql_error(c->db));

	return r;
    }
    r->res = mysql_store_result(c->db);
    //r->res = mysql_use_result(c->db);

    /* cleanup */
    for (it = q->sql_cleanup; it; it = it->next) {
        if (mysql_query(c->db, it->data)) {
	    r->error_str = g_strdup_printf("SQL error in '%s' %d: %s\n",
		    (char *)it->data, mysql_errno(c->db), mysql_error(c->db));

	    return r;
	}
    }

    for (i = 0; i< q->cols; i++) {
	q->bindings[i].sym = q->col_desc[i].sym;
    }

    r->row.col_desc = q->col_desc;
    r->row.bindings = q->bindings;
    r->row.cols = q->cols;

    return r;
}

ts_row *ts_result_fetch_row(ts_result *res)
{
    if (!res->res) {
	return NULL;
    }
    res->row.row = mysql_fetch_row(res->res);

    if (res->row.row) return &(res->row);

    return NULL;
}

ts_binding *ts_row_get_binding(ts_row *row, int column)
{
    ts_binding *b;

    if (column < 0 || column > row->cols) {
	return NULL;
    }

    b = row->bindings+column;
    b->lexical = row->row[row->col_desc[column].col];
    if (row->col_desc[column].dt_col >= 0) {
	b->datatype = row->row[row->col_desc[column].dt_col];
    } else {
	b->datatype = NULL;
    }
    if (row->col_desc[column].lang_col >= 0) {
	b->lang = row->row[row->col_desc[column].lang_col];
    } else {
	b->lang = NULL;
    }

    return b;
}

ts_binding *ts_row_get_binding_by_name(ts_row *row, const char *name)
{
    int column;

    for (column = 0; column < row->cols; column++) {
	if (!strcmp(row->col_desc[column].sym, name)) {
	    return ts_row_get_binding(row, column);
	}
    }

    return NULL;
}

void ts_result_free(ts_result *res)
{
    if (res) {
	mysql_free_result(res->res);
    }
    free(res);
}

int ts_query_complexity(ts_connection *c, ts_query *q)
{
    MYSQL_RES *res;
    MYSQL_ROW row;
    char *e_query;
    char *explain;
    int comp = 0;
    int need_free = 0;

    if (q->sql_pre_query) {
	e_query = q->sql_pre_query->data;
    } else {
	e_query = q->sql_query;
    }

    if (!e_query) {
	return 0;
    }

    if (!strncmp(e_query, "INSERT", 6)) {
	/* skip the first line of the prequery as its an INSERT */
	e_query = strchr((char *)(q->sql_pre_query->data), '\n') + 1;
    }

    if (!*e_query) {
	if (q->sql_pre_query->next) {
	    char *p;

	    e_query = strdup(strchr((char *)(q->sql_pre_query->next->data), '\n')+1);
	    need_free = 1;
	    p = strstr(e_query, "tmp0,");
	    if (p) {
		memcpy(p, "     ", 5);
	    }
	}
    }

    if (!*e_query) {
	fprintf(stderr, "Warning: cannot calculate complexity\n");

	return 0;
    }

    explain = g_strdup_printf("EXPLAIN %s", e_query);
    ts_mysql_query(c->db, explain);
    free(explain);
    if (need_free) {
	free(e_query);
    }
    res = mysql_use_result(c->db);
    if (!res) {
	fprintf(stderr, "Warning cannot calculate complexity\n");

	return 0;
    }
    while ((row = mysql_fetch_row(res))) {
	comp += atoi(row[8]);
    }

    if (q->sql_pre_query) {
	int base_comp = comp;

	/* Add on rows for the value joins */
	comp += q->cols * base_comp;

	/* Add on an estimate of the ciomplexity for the optionals */
	comp += base_comp * sqrt(g_slist_length(q->sql_pre_query));
    }

    return comp;
}

/* vi:set ts=8 sts=4 sw=4: */
