/*
 *  Copyright (C) 2005 Steve Harris
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  $Id: results.c,v 1.14 2005/12/21 14:29:09 theno23 Exp $
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <rasqal.h>
#include <glib.h>

#include "tstore.h"
#include "results.h"

void escape_node(const char *in, char *out);

void output_sparql(ts_result *res, ts_outp out, void *out_data, int headers, const char *stylesheet)
{
    ts_row *row;
    int i;
    GHashTable *bnodes = g_hash_table_new_full(g_str_hash, g_str_equal, NULL,
	    NULL);
    long bnodenum = 1;

    if (headers & TS_BIT_HEADERS) {
	out(out_data, "Content-Type: text/xml\r\n\r\n");
    }
    out(out_data, "<?xml version=\"1.0\"?>\n");
    if (stylesheet) {
	out(out_data, "<?xml-stylesheet type=\"text/xsl\" href=\"%s\" ?>\n",
	       stylesheet);
    }

    out(out_data, "<sparql xmlns=\"http://www.w3.org/2005/sparql-results#\">\n");
    out(out_data, "  <head>\n");
    if (res->error_str) {
	out(out_data, "    <error><![CDATA[\n%s    ]]></error>\n", res->error_str);
	out(out_data, "  </head>\n</sparql>\n");

	return;
    }
	
    for (i = 0; i < res->row.cols; i++) {
	out(out_data, "    <variable name=\"%s\"/>\n", res->row.col_desc[i].sym);
    }
    out(out_data, "  </head>\n");

    if (rasqal_query_get_verb(res->query->rquery) == RASQAL_QUERY_VERB_ASK) {
	row = ts_result_fetch_row(res);
	out(out_data, "  <boolean>%s</boolean>\n", row ? "true" : "false");
	out(out_data, "  </results>\n</sparql>\n");

	return;
    }

    out(out_data, "  <results ordered=\"%s\" distinct=\"%s\">\n",
	res->query->ordered ? "true" : "false",
	res->query->distinct ? "true" : "false");
    while ((row = ts_result_fetch_row(res))) {
	out(out_data, "    <result>\n");
	for (i = 0; i < row->cols; i++) {
	    ts_binding *b = ts_row_get_binding(row, i);
	    gchar *text;

	    if (b->lexical) {
		int dt = b->datatype ? atoi(b->datatype) : res->con->uri_dt;
		int lang = b->lang ? atoi(b->lang) : 0;

		text = g_markup_escape_text(b->lexical, strlen(b->lexical));

		if (!dt) {
		    if (lang) {
			char *lang_str = ts_get_language(res->con, lang);

			out(out_data, "      <binding name=\"%s\" xml:lang=\"%s\"><literal>%s</literal></binding>\n", b->sym, lang_str, text);
			free(lang_str);
		    } else {
			out(out_data, "      <binding name=\"%s\"><literal>%s</literal></binding>\n", b->sym, text);
		    }
		} else if (dt == res->con->uri_dt) {
		    if (*b->lexical == '_') {
			long num = (long)g_hash_table_lookup(bnodes, b->lexical);
			if (!num) {
			    g_hash_table_insert(bnodes, (void *)b->lexical,
				    (void *)bnodenum);
			    num = bnodenum++;
			}
			out(out_data, "      <binding name=\"%s\"><bnode>b%ld</bnode></binding>\n", b->sym, num);
		    } else {
			out(out_data, "      <binding name=\"%s\"><uri>%s</uri></binding>\n", b->sym, text);
		    }
		} else {
		    char *type = ts_get_datatype(res->con, dt);

		    out(out_data, "      <binding name=\"%s\"><literal datatype=\"%s\">%s</literal></binding>\n", b->sym, type, text);
		    free(type);
		}
		g_free(text);
	    } else {
		out(out_data, "      <binding name=\"%s\"><unbound /></binding>\n",
		       b->sym);
	    }
	}
	out(out_data, "    </result>\n");
    }
    out(out_data, "  </results>\n</sparql>\n");
}

static raptor_identifier_type slot_fill(void **data, raptor_statement *st,
	rasqal_literal *l, ts_result *res, ts_row *row, char tmp[32],
	GHashTable *bnodes, long *bnodenum)
{
    ts_binding *b;

    switch (l->type) {
    case RASQAL_LITERAL_URI:
	*data = l->value.uri;

	return RAPTOR_IDENTIFIER_TYPE_RESOURCE;

    case RASQAL_LITERAL_BLANK:
	*data = l->value.uri;

	return RAPTOR_IDENTIFIER_TYPE_ANONYMOUS;

    case RASQAL_LITERAL_STRING:
    case RASQAL_LITERAL_BOOLEAN:
    case RASQAL_LITERAL_INTEGER:
    case RASQAL_LITERAL_DOUBLE:
    case RASQAL_LITERAL_FLOAT:
    case RASQAL_LITERAL_DECIMAL:
    case RASQAL_LITERAL_DATETIME:
	*data = (void *)l->string;

	return RAPTOR_IDENTIFIER_TYPE_LITERAL;

    case RASQAL_LITERAL_VARIABLE:
	b = ts_row_get_binding_by_name(row, (char *)l->value.variable->name);
	if (b && b->lexical) {
	    int dt = b->datatype ? atoi(b->datatype) : res->con->uri_dt;
	    int lang = b->lang ? atoi(b->lang) : 0;

	    if (!dt) {
		if (lang) {
		    char *lang_str = ts_get_language(res->con, lang);

		    st->object_literal_language = (unsigned char *)lang_str;
		}
		*data = (void *)b->lexical;

		return RAPTOR_IDENTIFIER_TYPE_LITERAL;
	    } else if (dt == res->con->uri_dt) {
		if (*b->lexical == '_') {
		    long num = (long)g_hash_table_lookup(bnodes, b->lexical);
		    if (num) {
			snprintf(tmp, 32, "b%ld", num);
		    } else {
			g_hash_table_insert(bnodes, (void *)b->lexical,
				(void *)*bnodenum);
			snprintf(tmp, 32, "b%ld", *bnodenum);
			(*bnodenum)++;
		    }
		    *data = tmp;

		    return RAPTOR_IDENTIFIER_TYPE_ANONYMOUS;
		}
		*data = (void *)b->lexical;

		return RAPTOR_IDENTIFIER_TYPE_RESOURCE;
	    } else {
		*data = (void *)b->lexical;

		return RAPTOR_IDENTIFIER_TYPE_LITERAL;
	    }
	}
	*data = NULL;

	return RAPTOR_IDENTIFIER_TYPE_UNKNOWN;

    /* this should never happen */
    case RASQAL_LITERAL_PATTERN:
    case RASQAL_LITERAL_QNAME:
    case RASQAL_LITERAL_UNKNOWN:
	break;
    }

    *data = NULL;

    return RAPTOR_IDENTIFIER_TYPE_UNKNOWN;
}

void output_construct_graph(ts_result *res, ts_outp out, void *out_data, int headers, const char *stylesheet)
{
    ts_row *row;
    raptor_serializer *s = raptor_new_serializer("rdfxml");
    raptor_statement *st = calloc((sizeof(raptor_statement)), 1);
    GHashTable *bnodes = g_hash_table_new_full(g_str_hash, g_str_equal, NULL,
	    NULL);
    char tmp[32];
    long bnodenum = 1;
    int i;

    if (headers) {
	out(out_data, "Content-Type: application/rdf+xml\r\n\r\n");
    }

    for (i=0; 1; i++) {
	rasqal_prefix *p = rasqal_query_get_prefix(res->query->rquery, i);

	if (!p) {
	    break;
	}
	raptor_serialize_set_namespace(s, p->uri, p->prefix);
    }

    raptor_serialize_start_to_file_handle(s, res->query->base, stdout);
    while ((row = ts_result_fetch_row(res))) {
	for (i=0; 1; i++) {
	    rasqal_triple *trip =
		rasqal_query_get_construct_triple(res->query->rquery, i);

	    if (!trip) break;

	    st->subject_type =
		slot_fill((void **)&st->subject, st, trip->subject, res, row,
			tmp, bnodes, &bnodenum);
	    st->predicate_type =
		slot_fill((void **)&st->predicate, st, trip->predicate, res,
			row, tmp, bnodes, &bnodenum);
	    st->object_type =
		slot_fill((void **)&st->object, st, trip->object, res, row,
			tmp, bnodes, &bnodenum);
	    raptor_serialize_statement(s, st);
	}
    }
    free(st);
    g_hash_table_destroy(bnodes);
    raptor_serialize_end(s);
    raptor_free_serializer(s);
}

void output_testcase(ts_result *res, ts_outp out, void *out_data, int headers, const char *stylesheet)
{
    ts_row *row;
    int i, index, count;
    GHashTable *bnodes = g_hash_table_new_full(g_str_hash, g_str_equal, NULL, NULL);
    long bnodenum = 1;

    if (headers) {
	out(out_data, "Content-Type: application/rdf+turtle\r\n\r\n");
    }
    out(out_data, "@prefix rs: <http://www.w3.org/2001/sw/DataAccess/tests/result-set#> .\n");
    out(out_data, "@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .\n");
    out(out_data, "@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .\n");
    out(out_data, "\n[] rdf:type rs:ResultSet");
    if (res->row.cols > 0) {
    out(out_data, " ;\n   rs:resultVariable ");
	for (i = 0; i < res->row.cols; i++) {
	    if (i > 0) {
		out(out_data, ", ");
	    }
	    out(out_data, "\"%s\"", res->row.col_desc[i].sym);
	}
    }

    if (res->error_str) {
	out(out_data, " ;\n    rs:comment \"%s\"", res->error_str);
    }

    index = 0;
    while ((row = ts_result_fetch_row(res))) {
	out(out_data, " ;\n");
	out(out_data, "   rs:solution [\n");
	if (res->query->ordered) {
	    out(out_data, "      rs:index %d ;\n", ++index);
	}

	count = 0;
	for (i = 0; i < row->cols; i++) {
	    ts_binding *b = ts_row_get_binding(row, i);

	    if (b->lexical) {
		int dt = b->datatype ? atoi(b->datatype) : res->con->uri_dt;
		int lang = b->lang ? atoi(b->lang) : 0;

		if (count++ > 0) {
		    out(out_data, " ;\n");
		}
		out(out_data, "      rs:binding [ rs:variable \"%s\" ;\n", b->sym);
		if (!dt) {
		    if (lang) {
			char *p;
			char *lang_str = ts_get_language(res->con, lang);

			for (p=lang_str; *p; p++) {
			    *p=tolower(*p);
			}

			out(out_data, "                   rs:value \"%s\"@%s\n", b->lexical, lang_str);
			free(lang_str);
		    } else {
			out(out_data, "                   rs:value \"%s\"\n", b->lexical);
		    }
		} else if (dt == res->con->uri_dt) {
		    if (*b->lexical == '_') {
			long num = (long)g_hash_table_lookup(bnodes, b->lexical);
			if (num) {
			    out(out_data, "                   rs:value _:b%lds\n", num);
			} else {
			    g_hash_table_insert(bnodes, (void *)b->lexical,
				    (void *)bnodenum);
			    out(out_data, "                   rs:value _:b%ld\n",
				   bnodenum++);
			}
		    } else {
			out(out_data, "                   rs:value <%s>\n",
			       b->lexical);
		    }
		} else if (dt == res->con->int_dt) {
		    out(out_data, "                   rs:value %s\n", b->lexical);
		} else if ( dt == res->con->double_dt) {
		    if (strchr(b->lexical, '.')) {
			out(out_data, "                   rs:value %s\n", b->lexical);
		    } else {
			out(out_data, "                   rs:value \"%s\"^^xsd:double\n", b->lexical);
		    }
		} else if (dt == res->con->float_dt) {
		    out(out_data, "                   rs:value \"%s\"^^xsd:float\n", b->lexical);
		} else if (dt == res->con->bool_dt) {
		    out(out_data, "                   rs:value \"%s\"^^xsd:boolean\n", b->lexical);
		} else {
		    char *type = ts_get_datatype(res->con, dt);

		    out(out_data, "                   rs:value \"%s\"^^<%s>\n", 
                           b->lexical, type);
		    free(type);
		}
		out(out_data, "      ]");
	    }
	}
	out(out_data, "\n   ]");
    }
    out(out_data, " .\n");

    g_hash_table_destroy(bnodes);
}

void output_v2(ts_result *res, ts_outp out, void *out_data, int headers, const char *stylesheet)
{
    //XXX
    out(out_data, "!!!unsupported!!!\n");
}

void output_ascii(ts_result *res, ts_outp out, void *out_data, int headers, const char *stylesheet)
{
    ts_row *row;
    int i, index, count;
    GHashTable *bnodes = g_hash_table_new_full(g_str_hash, g_str_equal, NULL, NULL);
    long bnodenum = 1;
    int escbuflen = 1024;
    char *escbuf = malloc(escbuflen);

    if (headers) {
	out(out_data, "Content-Type: text/plain\r\n\r\n");
    }
    if (res->error_str) {
	char *pos = res->error_str;
	char *nextpos = pos;

	while (nextpos && *pos != '\0') {
	    nextpos = strchr(pos, '\n');
	    if (nextpos) {
		*nextpos = '\0';
		out(out_data, "# %s\n", pos);
		pos = nextpos + 1;
	    }
	}
    }
    if (res->row.cols > 0) {
	for (i = 0; i < res->row.cols; i++) {
	    if (i > 0) {
		out(out_data, "\t");
	    }
	    out(out_data, "%s", res->row.col_desc[i].sym);
	}
	out(out_data, "\n");
    }

    index = 0;
    while ((row = ts_result_fetch_row(res))) {
	count = 0;
	for (i = 0; i < row->cols; i++) {
	    ts_binding *b = ts_row_get_binding(row, i);

	    if (b->lexical) {
		int dt = b->datatype ? atoi(b->datatype) : res->con->uri_dt;
		int lang = b->lang ? atoi(b->lang) : 0;
 
		if (escbuflen < strlen(b->lexical) * 2 + 1) {
		    while (escbuflen < strlen(b->lexical) * 2 + 1) {
			escbuflen *= 2;
		    }
		    free(escbuf);
		    escbuf = malloc(escbuflen);
		}
		escape_node(b->lexical, escbuf);

		if (count++ > 0) {
		    out(out_data, "\t");
		}
		if (!dt) {
		    if (lang) {
			char *p;
			char *lang_str = ts_get_language(res->con, lang);

			for (p=lang_str; *p; p++) {
			    *p=tolower(*p);
			}

			out(out_data, "\"%s\"@%s", escbuf, lang_str);
			free(lang_str);
		    } else {
			out(out_data, "\"%s\"", escbuf);
		    }
		} else if (dt == res->con->uri_dt) {
		    if (*b->lexical == '_') {
			long num = (long)g_hash_table_lookup(bnodes, b->lexical);
			if (num) {
			    out(out_data, "_:b%lds", num);
			} else {
			    g_hash_table_insert(bnodes, (void *)b->lexical,
				    (void *)bnodenum);
			    out(out_data, "_:b%ld", bnodenum++);
			}
		    } else {
			out(out_data, "<%s>", escbuf);
		    }
		} else {
		    char *type = ts_get_datatype(res->con, dt);

		    out(out_data, "\"%s\"^^<%s>", escbuf, type);
		    free(type);
		}
	    }
	}
	out(out_data, "\n");
    }

    g_hash_table_destroy(bnodes);
}

void output_json(ts_result *res, ts_outp out, void *out_data, int headers, const char *stylesheet)
{
    ts_row *row;
    int i, index, count;
    int escbuflen = 1024;
    char *escbuf = malloc(escbuflen);

    if (headers) {
	out(out_data, "Content-Type: application/sparql-results+json\r\n\r\n");
    }
    out(out_data, "{\n  'head': {\n");
    if (res->error_str) {
	out(out_data, "    'error': [ '%s' ] },\n", res->error_str);
    }
    out(out_data, "    'vars': [ ");
    for (i = 0; i < res->row.cols; i++) {
	out(out_data, "'%s'", res->row.col_desc[i].sym);
	if (i < res->row.cols - 1) {
	    out(out_data, ", ");
	}
    }
    out(out_data, " ] \n  },\n  'results': {\n");
    out(out_data, "    'distinct': %s, ", res->query->distinct ? "true" : "false");
    out(out_data, "'ordered': %s,\n", res->query->ordered ? "true" : "false");
    out(out_data, "    'bindings': [\n     ");

    index = 0;
    while ((row = ts_result_fetch_row(res))) {
	count = 0;
	if (index++ > 0) {
	    out(out_data, ",");
	}
	out(out_data, " {\n");
	for (i = 0; i < row->cols; i++) {
	    ts_binding *b = ts_row_get_binding(row, i);

	    if (i > 0) {
		out(out_data, ",\n");
	    }
	    out(out_data, "        '%s': {", b->sym);
	    if (b->lexical) {
		int dt = b->datatype ? atoi(b->datatype) : res->con->uri_dt;
		int lang = b->lang ? atoi(b->lang) : 0;
 
		if (escbuflen < strlen(b->lexical) * 2 + 1) {
		    while (escbuflen < strlen(b->lexical) * 2 + 1) {
			escbuflen *= 2;
		    }
		    free(escbuf);
		    escbuf = malloc(escbuflen);
		}
		escape_node(b->lexical, escbuf);

		if (count++ > 0) {
		    out(out_data, ",");
		}
		out(out_data, "'value': \"%s\"", b->lexical);
		if (lang) {
		    char *p;
		    char *lang_str = ts_get_language(res->con, lang);

		    for (p=lang_str; *p; p++) {
			*p=tolower(*p);
		    }

		    out(out_data, ", 'xml:lang': \"%s\"", lang_str);
		    free(lang_str);
		}
		if (dt == res->con->uri_dt) {
		    if (*b->lexical == '_') {
			out(out_data, ", 'type': \"bnode\"");
		    } else {
			out(out_data, ", 'type': \"uri\"");
		    }
		} else {
		    char *type = ts_get_datatype(res->con, dt);

		    out(out_data, ", 'datatype': \"%s\"", type);
		    free(type);
		}
	    } else {
		out(out_data, "'type': \"unbound\"");
	    }
	    out(out_data, " }");
	}
	out(out_data, "\n      }");
    }
    out(out_data, "\n    ]\n  }\n}\n");
}

void escape_node(const char *in, char *out)
{
    while (*in) {
	char c = *in++;

	switch (c) {
	case '\\':
	case '"':
	case '\'':
	    *out++ = '\\';
	    break;
	case '\r':
	    *out++ = '\\';
	    c = 'r';
	    break;
	case '\n':
	    *out++ = '\\';
	    c = 'n';
	    break;
	case '\t':
	    *out++ = '\\';
	    c = 't';
	    break;
	}
	*out++ = c;
    }

    *out = '\0';
}

typedef void (*ts_ser_func)(ts_result *res, ts_outp out, void *out_data,
			    int headers, const char *stylesheet);

static const struct { const char *label; ts_ser_func func; } ser_functions[] = {
    {"ascii",    output_ascii},
    {"json",     output_json},
    {"sparql",   output_sparql},
    {"testcase", output_testcase},
    {"3store2",  output_v2},
    {NULL,       NULL}
};

void ts_serialise_results(ts_result *res, const char *syntax, ts_outp out, 
			  void *out_data, int flags, const char *style)
{
    if (res->query && res->query->rquery &&
	rasqal_query_get_verb(res->query->rquery) ==
	RASQAL_QUERY_VERB_CONSTRUCT) {
        output_construct_graph(res, out, out_data, flags, NULL);
    } else {
	int i;

	for (i=0; ser_functions[i].label; i++) {
	    if (!strcmp(syntax, ser_functions[i].label)) {
		ser_functions[i].func(res, out, out_data, flags, style);
		break;
	    }
	}
	if (!ser_functions[i].label) {
	    if (flags && TS_BIT_HEADERS) {
		out(out_data, "Content-Type: text/plain\r\n\r\n");
	    }
	    out(out_data, "error: unknown serialisation format '%s'\n",
		    syntax);
	}
    }
}

const char *ts_serialisation_format_by_index(int index)
{
    return ser_functions[index].label;
}

/* vi:set ts=8 sts=4 sw=4: */
