/*
 *  Copyright (C) 2005 Steve Harris
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  $Id: sql-compile.c,v 1.30 2006/01/02 21:36:57 theno23 Exp $
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <raptor.h>
#include <rasqal.h>
#include <glib.h>
#include <unistd.h>
#include <sys/types.h>

#include "tstore.h"
#include "config.h"

//#define g_string_append(a, b) {if(!b) fprintf(stderr, "NULL param at %s:%d\n", __FILE__, __LINE__); g_string_append(a, b); }

#define TMP_SIZE 256

/* define what column is used for joins between the triples and symbols
 * tables */
#ifdef ENABLE_QUADS
#define IDC "hash"
#else
#define IDC "sid"
#endif

/* walk over a triple list and bind all the variables as appropriate */
static int ts_parse_graph_pattern(ts_symtab *symtab,
    rasqal_graph_pattern *gp, GSList *sc[], int *opt_block, int table_block[],
    int reqd, int opt);

/* walk the constraint list and handle each expression tree */
static void ts_parse_constr_seq(ts_symtab *symtab, raptor_sequence *seq,
    GSList **sc, int block);

/* bind individual triple slots (subject, predicate etc.) */
static void ts_check_triple_slot(ts_symtab *symtab, const char *table,
    rasqal_literal *l, const char *slot, GSList **sc, int block);

/* walk down an expression tree and build SQL expressions representing the
 * SPARQL expressions */
static void ts_expression_descend(ts_symtab *st, rasqal_expression *e,
    GString *s, rasqal_literal_type prom, int block);

/* finds wether an given expression needs to be bound late, ie. after all the sub graphs have been built into tmp tables and joined */
static int ts_expression_latebind(ts_symtab *st, rasqal_expression *e);

/* maps SPARQL operators into SQL ones */
static const char *ts_sql_operator(rasqal_op op);

/* something iffy around type promotion, not SPARQL complaint */
static rasqal_literal_type ts_expression_promoted_type(rasqal_literal_type t,
    rasqal_expression *e);

/* is this operator a boolean predicate? */
static int ts_boolean_op(rasqal_expression *e);

/* is this operator a binary numerical operator? eg. (?x) + (?y) */
static int ts_binary_numerical_op(rasqal_expression *e);

/* build the SQL expression to make a tmp table */
static char *make_tmp_table(raptor_sequence *symbols, ts_symtab *st, int block,
		       int distinct, int table_block[], GSList *sql_constrs[]);

/* build the SQL expression to populate a tmp table */
static ts_rel_query *fill_tmp_table(raptor_sequence *symbols, ts_symtab *symtab,
			   int block, int distinct, int limit,
			   int table_block[], GSList *sql_constrs[]);

/* add a set of constraints to s that restricts the expression e to be of a
 * known type if it's a vanilla variable */
static void add_known_type_constraint(GString *s, rasqal_expression *e,
					ts_symtab *st, int block);

/* utility function to covert a symtab into a sequence */
void seq_build(gpointer key, gpointer value, gpointer user_data);

/* turn tmp table filling sintructions into a subselect */
static char *subselect_first_prequery(ts_query *q, GSList **pq, int tbl, int skip);

static void slot_check(GHashTable *used, raptor_sequence *s, rasqal_literal *l);

/* erase some symtab members for variables that are only in unions */
void union_clear(gpointer key, gpointer value, gpointer user_data);

/* return the table name for an internal type index */
static const char *type_table_name(int type);

void graph_pattern_walk(rasqal_graph_pattern *pattern, int *i, int *opt_block,
	int table_block[], ts_symtab *symtab, GSList *sql_constrs[], int reqd,
	int opt, int uni)
{
    int index = 0;
    int flags = rasqal_graph_pattern_get_operator(pattern);

    if (flags == RASQAL_GRAPH_PATTERN_OPERATOR_OPTIONAL) {
	if (!reqd) {
	    (*opt_block)++;
	}
	opt++;
    }

    if (flags == RASQAL_GRAPH_PATTERN_OPERATOR_UNION) {
	uni++;
	opt++;
	if (!reqd) symtab->unions++;
    }

    /* we do this processing in an odd order:
     *
     * 1) the current graph patterns, just so that the binding order isn't
     *    confusing when you look at the SQL output
     * 2) the sub graph patters, to make sure that all the pattern variables
     *     are bound
     * 3) the constraints, so that it can apply them as early as possible
     */
    if ((opt && !reqd) || (!opt && reqd)) {
	ts_parse_graph_pattern(symtab, pattern, sql_constrs, opt_block,
			   table_block, reqd, opt);
    }
    while (!(opt && reqd)) {
	rasqal_graph_pattern *sgp =
	    rasqal_graph_pattern_get_sub_graph_pattern(pattern, index);
	if (!sgp) {
	    break;
	}
	if (uni && rasqal_graph_pattern_get_sub_graph_pattern(pattern, 0)) {
	    if (!reqd) {
		(*opt_block)++;
		symtab->union_group[*opt_block] = symtab->unions;
// printf("... putting block %d into group %d\n", *opt_block, symtab->unions);
	    }
	}

	(*i)++;
	index++;
	graph_pattern_walk(sgp, i, opt_block, table_block, symtab, 
		           sql_constrs, reqd, opt, uni);
	if (uni) {
	    /* unbind any variables that are bound only in UNION blocks,
	     * they are effectivly scoped to those blocks */
	    g_hash_table_foreach(symtab->hash, union_clear, opt_block);
	}
    }
    if ((opt && !reqd) || (!opt && reqd)) {
	ts_parse_constr_seq(symtab,
			rasqal_graph_pattern_get_constraint_sequence(pattern),
			&sql_constrs[*opt_block], *opt_block);
    }
}

void union_clear(gpointer key, gpointer value, gpointer user_data)
{
    ts_symbol *s = value;

    if (s->block > 0 && s->st->union_group[s->block] && s->hash_col) {
	//XXX memory leak?
	s->last_hash[s->block] = s->hash_col;
	s->hash_col = NULL;
	s->tt_col = NULL;
    }
}

static int one_bit_set(unsigned int bm)
{
    do {
	if (bm == 1) {
	    return 1;
	}
	/* if the bottom bit is set, but so are more, 2+ bits are set */
	if ((bm & 1) == 1) {
	    return 0;
	}
	bm = bm >> 1;
    } while (bm > 0);

    return 0;
}

static int bitmask_valid(ts_symtab *symtab, int blocks, unsigned int bm)
{
    int group, block;

    for (group=1; group<=symtab->unions; group++) {
	unsigned int gmask = 0;

	for (block=1; block<=blocks; block++) {
	    if (symtab->union_group[block] == group) {
		gmask |= (1 << (block - 1));
	    }
	}

	/* each group must only have one block in any mask */
	if (!one_bit_set(bm & gmask)) {
	    return 0;
	}
    }

    return 1;
}

ts_query *ts_sql_compile(ts_query *q, ts_connection *c, rasqal_query *rq)
{
    raptor_sequence *vars = rasqal_query_get_bound_variable_sequence(rq);
    raptor_sequence *symbols;
    rasqal_graph_pattern *pattern;
    rasqal_query_verb verb;
    char *tmp_t;
    ts_rel_query *tmp_tq;
    int count;
    int i, j, opt;
    int union_blocks = 0;
    const int distinct = rasqal_query_get_distinct(rq);
    int limit = rasqal_query_get_limit(rq);
    ts_symtab *symtab;
    int opt_block = 0;
    GSList *sql_constrs[TS_MAX_BLOCKS], *it;
    GSList *pre_query_rel = NULL;
    int table_block[TS_MAX_BLOCKS];
    GString *ret;
    GString *con, *constr;
    GString *order = g_string_new("");

    if (!q) {
	q = calloc(1, sizeof(ts_query));
    }
// rasqal_query_print(rq, stdout);

    verb = rasqal_query_get_verb(rq);

    if (rasqal_query_get_order_condition(rq, 0)) {
	q->ordered = 1;
    }

    /* find any CONSTRUCTed variables, and copy them to the returned vars
     * structure */
    if (verb == RASQAL_QUERY_VERB_CONSTRUCT) {
	vars = raptor_new_sequence(NULL, NULL);
	GHashTable *used = g_hash_table_new(g_str_hash, g_str_equal);

	for (i=0; 1; i++) {
	    rasqal_triple *trip = rasqal_query_get_construct_triple(rq, i);
	    if (!trip) break;

	    slot_check(used, vars, trip->subject);
	    slot_check(used, vars, trip->predicate);
	    slot_check(used, vars, trip->object);
	}
	g_hash_table_destroy(used);
    }

    if (!vars) {
	count = 0;
	limit = 1;
    } else {
	count = raptor_sequence_size(vars);
    }
    q->vars = vars;
    symtab = ts_symtab_new(c, q);
    q->symtab = symtab;

    for (i=0; i<TS_MAX_BLOCKS; i++) {
	sql_constrs[i] = NULL;
    }

    /* all these are SELECTed */
    for (i=0; i < count; i++) {
	rasqal_variable *v = raptor_sequence_get_at(vars, i);

	ts_symtab_get(symtab, v->name, FALSE, 0);
    }
// ts_symtab_dump(symtab, stderr);

    pattern = rasqal_query_get_query_graph_pattern(rq);
    if (!pattern) {
	q->error_str = g_strdup("No query graph pattern found");

	return q;
    }

    i = 0;
    graph_pattern_walk(pattern, &i, &opt_block, table_block, symtab, sql_constrs, TRUE, 0, 0);

    symbols = raptor_new_sequence(NULL, NULL);
    g_hash_table_foreach(symtab->hash, seq_build, symbols);
// ts_symtab_dump(symtab, stdout);

    /* constrain the joins between triple and symbol tables */
    for (i=0; i < raptor_sequence_size(symbols); i++) {
	ts_symbol *s = raptor_sequence_get_at(symbols, i);
	char tmp[TMP_SIZE];

	if (s->hash_col && ts_symbol_valtbl(s, 0)) {
	    snprintf(tmp, TMP_SIZE, "%s=%s."IDC, s->hash_col, ts_symbol_valtbl(s, 0));
	    sql_constrs[0] = g_slist_prepend(sql_constrs[0], strdup(tmp));
	}
    }

    tmp_t = make_tmp_table(symbols, symtab, 0, distinct, table_block,
			   sql_constrs);
    q->sql_setup = g_slist_append(q->sql_setup, tmp_t);
    tmp_tq = fill_tmp_table(symbols, symtab, 0, distinct, limit, table_block,
			   sql_constrs);
    pre_query_rel = g_slist_append(pre_query_rel, tmp_tq);
    q->sql_cleanup = g_slist_append(q->sql_cleanup,
			     g_strdup_printf("DROP TABLE %s", q->tt_names[0]));

    i = 0;
    graph_pattern_walk(pattern, &i, &opt_block, table_block, symtab, sql_constrs, FALSE, 0, 0);

    raptor_free_sequence(symbols);
    symbols = raptor_new_sequence(NULL, NULL);
    g_hash_table_foreach(symtab->hash, seq_build, symbols);

    for (opt=1; opt <= opt_block; opt++) {
	tmp_t = make_tmp_table(symbols, symtab, opt, distinct, table_block,
			       sql_constrs);
	q->sql_setup = g_slist_append(q->sql_setup, tmp_t);
	tmp_tq = fill_tmp_table(symbols, symtab, opt, distinct, limit,
			       table_block, sql_constrs);
	pre_query_rel = g_slist_append(pre_query_rel, tmp_tq);
	q->sql_cleanup = g_slist_append(q->sql_cleanup,
			 g_strdup_printf("DROP TABLE %s", q->tt_names[opt]));
	if (symtab->union_group[opt]) {
	    union_blocks++;
	}
    }

    if (symtab->unions) {
	GString *it_ins = g_string_new("INSERT INTO uit VALUES");
	const int expon = (1 << union_blocks);
	int count = 0;

	q->sql_setup = g_slist_append(q->sql_setup, strdup("CREATE TEMPORARY TABLE uit (bm integer NOT NULL)\n"));
	q->sql_cleanup = g_slist_append(q->sql_cleanup, strdup("DROP TABLE uit"));
	for (i=0; i < expon; i++) {
	    if (!bitmask_valid(symtab, opt_block, i)) {
		continue;
	    }
	    if (count++ > 0) {
		g_string_sprintfa(it_ins, ",(%d)", i);
	    } else {
		g_string_sprintfa(it_ins, "(%d)", i);
	    }
	}
	g_string_append(it_ins, "\n");
	q->sql_pre_query = g_slist_append(q->sql_pre_query, it_ins->str);
	g_string_free(it_ins, FALSE);
    }

    for (i=0; i < count; i++) {
	rasqal_variable *v = raptor_sequence_get_at(vars, i);

	ts_symtab_get(symtab, v->name, TRUE, 0);
    }

    ret = g_string_new("SELECT ");
    if (distinct) {
	g_string_append(ret, "DISTINCT ");
    }
    for (i=0; i < count; i++) {
	rasqal_variable *v = raptor_sequence_get_at(vars, i);
	ts_symbol *s = ts_symtab_get(symtab, v->name, FALSE, 0);
	const char *vtbl = ts_symbol_valtbl(s, 0);

	if (s->block == -1) {
	    g_string_sprintfa(ret, "NULL AS `%s`, 0 AS %s_dt, 0 AS %s_lang",
			      v->name, v->name, v->name);
	} else if (s->bound_in_msp && c->opt_lvl > 1) {
	    g_string_sprintfa(ret, "%s.lexical AS `%s`", vtbl, v->name);
	} else {
	    g_string_sprintfa(ret, "%s.lexical AS `%s`, %s.datatype AS %s_dt, "
			       "%s.language AS %s_lang",
			  vtbl, v->name, vtbl, v->name, vtbl, v->name);
	}
	if (i < count - 1) {
	    g_string_append(ret, ", ");
	}
    }
    if (!vars) {
	 g_string_append(ret, "1 ");
    }

    /* find the ORDER statements and SQL-ise them */
    if (q->ordered) {
	g_string_append(order, "ORDER BY ");
	rasqal_expression *e;

	for (i=0; (e = rasqal_query_get_order_condition(rq, i)); i++) {
	    if (i) g_string_append(order, ", ");
	    ts_expression_descend(symtab, e, order, RASQAL_LITERAL_UNKNOWN, TS_ORDER_BLOCK);
	}
    }

    if (c->opt_lvl > 0 && opt == 1) {
	char *tmp = subselect_first_prequery(q, &pre_query_rel, 0, 0);

	g_string_sprintfa(ret, "\nFROM %s", tmp);
	free(tmp);
    } else {
	g_string_sprintfa(ret, "\nFROM %s", q->tt_names[0]);
    }
    if (symtab->unions) {
	g_string_append(ret, ", uit");
    }

    /* find symbol joins that are needed by SELECTd variables */
    for (i=0; vars && i<raptor_sequence_size(vars); i++) {
	rasqal_variable *v = raptor_sequence_get_at(vars, i);
	ts_symbol *s = ts_symtab_get(symtab, v->name, FALSE, 0);

	if (s->block == 0 && ts_symbol_valtbl(s, 0)) {
	    g_string_sprintfa(ret, ", symbols %s", ts_symbol_valtbl(s, 0));
	}
    }

///XXX old order place marker

    for (opt=1; opt <= opt_block; opt++) {
	char *tbl;

	con = g_string_new("");
	for (i=0; i<raptor_sequence_size(symbols); i++) {
	    ts_symbol *s = raptor_sequence_get_at(symbols, i);

	    if (ts_symbol_appears_in(s, opt) && opt != s->block) {
		if (con->len > 0) g_string_append(con, " && ");
		g_string_sprintfa(con, "%s.%s=%s.%s", q->tt_names[s->block],
				  s->sym, q->tt_names[opt], s->sym);
	    }
	}

	if (c->opt_lvl > 1) {
	    tbl = subselect_first_prequery(q, &(pre_query_rel->next), opt, 1);
	    ts_list_replace_str(q->sql_setup, "TEMPORARY", "/* TMP */");
	} else {
	    tbl = strdup(q->tt_names[opt]);
	}

	if (con->len && symtab->union_group[opt] == 0) {
	    g_string_sprintfa(ret, " LEFT JOIN %s ON %s", tbl, con->str);
	} else {
	    g_string_sprintfa(ret, " LEFT JOIN %s ON 1", tbl);
	    if (symtab->union_group[opt]) {
		g_string_sprintfa(ret, " && (uit.bm & %d)", 1 << (opt-1));
	    }
	}

	free(tbl);

	g_string_append(ret, "\n");
	g_string_free(con, TRUE);
    }

    /* find all the constraints that have to be applied globally and apply
     * them */
    constr = g_string_new("");
    for (it = q->globe; it; it = it->next) {
	GString *s = g_string_new("");
	rasqal_expression *e = it->data;

	ts_expression_descend(symtab, e, s, RASQAL_LITERAL_BOOLEAN, 0);
	if (constr->len > 0) g_string_append(constr, " && ");
	g_string_append(constr, s->str);
	g_string_free(s, TRUE);
    }

    con = g_string_new("");
    for (i=0; i<raptor_sequence_size(symbols); i++) {
	ts_symbol *s = raptor_sequence_get_at(symbols, i);

	if (s->block != 0) {
	    const char *vt = ts_symbol_valtbl_reqd(s, 0);

	    if (s->tt_col) {
		/* nice and easy, the variable is only bound in one place */
		g_string_sprintfa(con, " LEFT JOIN "T_S" %s ON %s=%s."IDC"\n", vt, s->tt_col, vt);
		for (j=0; j<TS_TYPE_TABLES; j++) {
		    const char *tname;

		    /* skip this value table if its not used in this block */
		    if (!s->val_tbl_type[j][0]) {
			continue;
		    }

		    tname = type_table_name(j);
		    g_string_sprintfa(con, " LEFT JOIN %s %s ON %s."IDC"=%s."
					   IDC"\n", tname,
				      s->val_tbl_type[j][0],
				      ts_symbol_valtbl(s, 0),
				      s->val_tbl_type[j][0]);
		}
	    } else if (s->block != -1) {
		/* not so easy, the variable is bound in 1+ UNION blocks */
		int j;
		int count = 0;

		g_string_sprintfa(con, " LEFT JOIN "T_S" %s ON (", vt);
		for (j=1; j<=opt_block; j++) {
		    if (s->appears_in[j] && symtab->union_group[j]) {
			if (count++ > 0) {
			    g_string_append(con, " || ");
			}
			g_string_sprintfa(con, "%s.%s=%s."IDC, q->tt_names[j],
					  s->sym, vt);
		    }
		}
		g_string_sprintfa(con, ")\n");
		s->hash_col = g_strdup_printf("%s."IDC, vt);
	    }
	}
    }
    g_string_sprintfa(ret, "%s", con->str);
    g_string_free(con, TRUE);

    /* find symbol joins that are needed by ORDER BY constraints only */
    for (i=0; i<raptor_sequence_size(symbols); i++) {
	ts_symbol *s = raptor_sequence_get_at(symbols, i);

	if (ts_symbol_valtbl(s, TS_ORDER_BLOCK) && !ts_symbol_valtbl(s, 0)) {
	    g_string_sprintfa(ret, ", "T_S" %s", ts_symbol_valtbl(s, TS_ORDER_BLOCK));
	}
	for (j=0; j<TS_TYPE_TABLES; j++) {
	    if (s->val_tbl_type[j][TS_ORDER_BLOCK] &&
		ts_symbol_valtbl(s, TS_ORDER_BLOCK)) {
		const char *tname = type_table_name(j);

		g_string_sprintfa(ret, " LEFT JOIN %s %s ON %s."IDC"=%s."
				   IDC, tname,
				  s->val_tbl_type[j][TS_ORDER_BLOCK],
				  ts_symbol_valtbl(s, TS_ORDER_BLOCK),
				  s->val_tbl_type[j][TS_ORDER_BLOCK]);
	    }
	}
    }
    g_string_append(ret, "\n");

    for (i=0; i < count; i++) {
	rasqal_variable *v = raptor_sequence_get_at(vars, i);
	ts_symbol *s = ts_symtab_get(symtab, v->name, FALSE, 0);
	if (s->block != 0) continue;
	if (constr->len > 0) g_string_append(constr, " && ");
	g_string_sprintfa(constr, "%s=%s."IDC, s->tt_col, ts_symbol_valtbl(s, 0));
    }

    for (i=0; i<raptor_sequence_size(symbols); i++) {
	ts_symbol *s = raptor_sequence_get_at(symbols, i);

	if (ts_symbol_valtbl(s, TS_ORDER_BLOCK) && !ts_symbol_valtbl(s, 0)) {
	    if (constr->len > 0) g_string_append(constr, " && ");
	    g_string_sprintfa(constr, "%s=%s."IDC, s->tt_col,
			      ts_symbol_valtbl(s, TS_ORDER_BLOCK));
	}
    }

    if (*(constr->str)) g_string_sprintfa(ret, "WHERE %s\n", constr->str);
    g_string_free(constr, TRUE);

    g_string_append(ret, order->str);
    g_string_free(order, TRUE);

    if (limit >= 0) {
	g_string_sprintfa(ret, "\nLIMIT %d", limit);
    }

    q->sql_query = ret->str;
    g_string_free(ret, FALSE);

    for (it = pre_query_rel; it; it = it->next) {
	q->sql_pre_query = g_slist_append(q->sql_pre_query, ts_rel_query_to_string(it->data, "", ""));
	ts_rel_query_free(it->data);
    }

    raptor_free_sequence(symbols);

    return q;
}

static const char *type_table_name(int type)
{
    switch (type) {
	case 0:
	    return T_I;
	case 1:
	    return T_F;
	case 2:
	    return T_A;
	default:
	    return "!ERROR!";
    }
}

static char *subselect_first_prequery(ts_query *q, GSList **pq, int tbl, int skip)
{
    char *ret;
    ts_rel_query *rq = (*pq)->data;
    void *old;
    GSList *it;
    GSList *oldhead;
    char tname[32];
    int i;

    free(rq->map);
    rq->map = NULL;
    snprintf(tname, 31, ") AS `%s`", q->tt_names[tbl]);
    ret = ts_rel_query_to_string(rq, "(", tname);
    oldhead = *pq;
    *pq = oldhead->next;
    ts_rel_query_free(oldhead->data);
    g_slist_free_1(oldhead);

    for (i=0, it=q->sql_setup; i<skip; i++) {
	it = it->next;
    }
    old = it->data;
    q->sql_setup = g_slist_remove(q->sql_setup, old);
    free(old);

    for (i=0, it=q->sql_cleanup; i<skip; i++) {
	it = it->next;
    }
    old = it->data;
    q->sql_cleanup = g_slist_remove(q->sql_cleanup, old);
    free(old);

    return ret;
}

static void slot_check(GHashTable *used, raptor_sequence *s, rasqal_literal *l)
{
    if (l->type == RASQAL_LITERAL_VARIABLE) {
	if (!g_hash_table_lookup(used, l->value.variable->name)) {
	    raptor_sequence_push(s, l->value.variable);
	    g_hash_table_insert(used, (char *)l->value.variable->name,
		    (void *)1);
	}
    }
}

static char *make_tmp_table(raptor_sequence *symbols, ts_symtab *symtab,
			   int block, int distinct, int table_block[],
			   GSList *sql_constrs[])
{
    GString *con;
    GString *ret;
    char *retstr;
    int i;

    ret = g_string_new("");
    con = g_string_new("");
    for (i=0; i < raptor_sequence_size(symbols); i++) {
	ts_symbol *s = raptor_sequence_get_at(symbols, i);
	if (!ts_symbol_appears_in(s, block)) {
	    continue;
	}
	if (con->len > 0) {
	    g_string_append(con, ",\n");
	}
	g_string_sprintfa(con, "   `%s` bigint NOT NULL, index(`%s`)", s->sym,
			  s->sym);
    }
    /* Incase this table isnt really needed, put a column in it so that it can
     * be joined. */
    if (con->len == 0) {
	g_string_append(con, "   foo int NOT NULL");
    }

    if (symtab->c->opt_lvl > 0) {
	//XXX should be a more robust unique number
	symtab->q->tt_names[block] = g_strdup_printf("tmp%d_%x", block, getpid());
    } else {
	symtab->q->tt_names[block] = g_strdup_printf("tmp%d", block);
    }
    g_string_sprintfa(ret, "CREATE TEMPORARY TABLE %s (\n%s\n)\n",
			symtab->q->tt_names[block], con->str);
    g_string_free(con, TRUE);

    retstr = ret->str;
    g_string_free(ret, FALSE);

    return retstr;
}

static ts_rel_query *fill_tmp_table(raptor_sequence *symbols, ts_symtab *symtab,
			   int block, int distinct, int limit,
			   int table_block[], GSList *sql_constrs[])
{
    GString *con;
    GSList *it;
    char tmp[TMP_SIZE];
    int i, j;
    ts_rel_query *ret = calloc(1, sizeof(ts_rel_query));

    con = g_string_new("");

    for (i=0; i<raptor_sequence_size(symbols); i++) {
	ts_symbol *s = raptor_sequence_get_at(symbols, i);

	if (!s->hash_col && !s->last_hash[block]) {
	    continue;
	}
	if (!ts_symbol_appears_in(s, block)) {
	    continue;
	}

	if (s->block == 0 || s->appears_in[block]) {
	    if (con->len > 0) {
		g_string_append(con, ", ");
	    }
	    if (!s->hash_col) {
		g_string_append(con, s->last_hash[block]);
	    } else {
		g_string_append(con, ts_symbol_hashcol(s));
	    }
	}

	if (!s->tt_col && !s->last_hash[block]) {
	    s->tt_col = g_strdup_printf("%s.%s", symtab->q->tt_names[block],
					s->sym);
	}

	g_string_sprintfa(con, " AS `%s`", s->sym);
    }

    if (con->len == 0) {
	g_string_free(con, TRUE);

	if (block) {
	    ret->map = g_strdup_printf("/* no data in table %d */\n", block);

	    return ret;
	} else {
	    ret->map = g_strdup_printf("INSERT INTO %s VALUES(0)\n",
				symtab->q->tt_names[0]);

	    return ret;
	}
    }	
    ret->map = g_strdup_printf("INSERT INTO %s\n", symtab->q->tt_names[block]);
    ret->project = g_strdup_printf("SELECT %s%s\n",
				   distinct ? "DISTINCT " : "", con->str);
    g_string_free(con, TRUE);

    con = g_string_new("");
    if (block > 0) {
	g_string_append(con, symtab->q->tt_names[0]);
    }
    for (i=0; i < symtab->hash_j_cnt; i++) {
	if (table_block[i] == block) {
	    if (con->len > 0) {
		g_string_append(con, ", ");
	    }
	    g_string_sprintfa(con, T_T" t%d", i);
#ifndef ENABLE_QUADS
	    if (symtab->model_joins[i]) {
		if (con->len > 0) {
		    g_string_append(con, ", ");
		}
		g_string_sprintfa(con, T_M" t%dm", i);
		snprintf(tmp, TMP_SIZE, "t%dm.tid=t%d.tid", i, i);
		sql_constrs[block] = g_slist_prepend(sql_constrs[block],
						     strdup(tmp));
	    }
#endif
	}
    }
    for (i=0; i<raptor_sequence_size(symbols); i++) {
	ts_symbol *s = raptor_sequence_get_at(symbols, i);

	if (ts_symbol_valtbl(s, block) && ts_symbol_appears_in(s, block)) {
	    if (con->len > 0) {
		g_string_append(con, ", ");
	    }
	    g_string_sprintfa(con, T_S" %s", ts_symbol_valtbl(s, block));
	    snprintf(tmp, TMP_SIZE, "%s."IDC"=%s", ts_symbol_valtbl(s, block),
		     s->hash_col);
	    sql_constrs[block] = g_slist_prepend(sql_constrs[block],
						 strdup(tmp));
	    for (j=0; j<TS_TYPE_TABLES; j++) {
		const char *tname;

		/* skip this value table if its not used in this block */
		if (!s->val_tbl_type[j][block]) {
		    continue;
		}

		tname = type_table_name(j);

		if (con->len > 0) {
		    g_string_append(con, ", ");
		}
		g_string_sprintfa(con, "%s %s", tname,
				  s->val_tbl_type[j][block]);
		snprintf(tmp, TMP_SIZE, "%s."IDC"=%s."IDC,
			 ts_symbol_valtbl(s, block), s->val_tbl_type[j][block]);
		sql_constrs[block] = g_slist_prepend(sql_constrs[block],
						     strdup(tmp));
	    }
	}
	/* XXX this is not quite right, as we only need to join the lang table
	 * if lang(?foo) is used in this block */
	if (s->lang_col && ts_symbol_appears_in(s, block)) {
	    if (con->len > 0) {
		g_string_append(con, ", ");
	    }
	    g_string_sprintfa(con, T_L" %s", s->lang_col);
	    snprintf(tmp, TMP_SIZE, "%s.id=%s.language", s->lang_col,
		     ts_symbol_valtbl(s, block));
	    sql_constrs[block] = g_slist_prepend(sql_constrs[block],
		    strdup(tmp));
	}
	if (s->dt_col && ts_symbol_appears_in(s, block)) {
	    if (con->len > 0) {
		g_string_append(con, ", ");
	    }
	    g_string_sprintfa(con, T_D" %s", s->dt_col);
	    snprintf(tmp, TMP_SIZE, "%s.id=%s.datatype", s->dt_col,
		     ts_symbol_valtbl(s, block));
	    sql_constrs[block] = g_slist_prepend(sql_constrs[block],
		    strdup(tmp));
	}
    }

    ret->join = g_strdup_printf("FROM %s\n", con->str);
    g_string_free(con, TRUE);

    con = g_string_new("");
    for (it=sql_constrs[block]; it; it=it->next) {
	if (it == sql_constrs[block]) {
	    g_string_append(con, "WHERE ");
	} else {
	    g_string_append(con, "AND ");
	}
	g_string_sprintfa(con, "%s\n", (char *)it->data);
    }

    if (limit > -1 && !symtab->q->ordered) {
	g_string_sprintfa(con, "\nLIMIT %d", limit);
    }

    ret->select = con->str;
    g_string_free(con, FALSE);

    return ret;
}

void seq_build(gpointer key, gpointer value, gpointer user_data)
{
    raptor_sequence *seq = user_data;

    raptor_sequence_push(seq, value);
}

static int ts_parse_graph_pattern(ts_symtab *symtab,
    rasqal_graph_pattern *gp, GSList *sc[], int *opt_block, int table_ob[],
    int reqd, int opt)
{
    rasqal_triple *t;
    int i;
    int handled;
    int idx;
    char table[TMP_SIZE];

    if (!gp) {
	return 0;
    }

    if (opt) {
	if (reqd) return 0;
	idx = *opt_block;
    } else {
	if (!reqd) return 0;
	idx = 0;
    }

    for (i=0; 1; i++) {
	t = rasqal_graph_pattern_get_triple(gp, i);
	if (!t) break;
	handled = 0;

	if (symtab->c->feature[TS_FEATURE_INFERENCE]) {
	    /* backchaining */
	    if (t->predicate->type == RASQAL_LITERAL_URI &&
		!strcmp(RDF_TYPE, (char *)raptor_uri_as_string(t->predicate->value.uri))) {
		rasqal_variable rv;
		rasqal_literal newvar;
		raptor_uri *ru;
		rasqal_literal scop;
		char vname[TMP_SIZE];

		handled = 1;

		/* (?a, <rdf:type>, <foo:bar>) =>
		 *   (?a, <rdf:type>, ?foo),
		 *   (?foo, <rdf:subClassOf>, <foo:bar>)
		 */

		newvar.type = RASQAL_LITERAL_VARIABLE;
		snprintf(vname, TMP_SIZE, "_TV%d", symtab->tmpvar_cnt++);
		rv.name = (unsigned char *)vname;
		newvar.value.variable = &rv;

		ru = raptor_new_uri((unsigned char *)RDFS_SUBCLASS);
		scop.type = RASQAL_LITERAL_URI;
		scop.value.uri = ru;

		table_ob[symtab->hash_j_cnt] = idx;
		snprintf(table, TMP_SIZE, "t%d", symtab->hash_j_cnt++);
		ts_check_triple_slot(symtab, table, t->origin, "model", &sc[idx], idx);
		ts_check_triple_slot(symtab, table, t->subject, "subject", &sc[idx], idx);
		ts_check_triple_slot(symtab, table, t->predicate, "predicate", &sc[idx], idx);
		ts_check_triple_slot(symtab, table, &newvar, "object", &sc[idx], idx);
		table_ob[symtab->hash_j_cnt] = idx;
		snprintf(table, TMP_SIZE, "t%d", symtab->hash_j_cnt++);
		ts_check_triple_slot(symtab, table, t->origin, "model", &sc[idx], idx);
		ts_check_triple_slot(symtab, table, &newvar, "subject", &sc[idx], idx);
		ts_check_triple_slot(symtab, table, &scop, "predicate", &sc[idx], idx);
		ts_check_triple_slot(symtab, table, t->object, "object", &sc[idx], idx);
		
	    } else if (t->predicate->type == RASQAL_LITERAL_URI &&
		!strcmp(DIRECT_TYPE, (char *)raptor_uri_as_string(t->predicate->value.uri))) {
		/* replace instances of direct:type with the raw rdf:type URI */
		t->predicate->value.uri = 
		    raptor_new_uri((unsigned char *)DIRECT_TYPE);
	    } else if (t->predicate->type == RASQAL_LITERAL_URI) {
		char *puri = (char *)raptor_uri_to_string(t->predicate->value.uri);
		ts_hash phash = ts_hash_uri(puri);
		ts_hash rspo = ts_hash_uri(RDFS_SUBPROPERTY);
		char cq[TMP_SIZE];
		int num;

		snprintf(cq, TMP_SIZE, "SELECT 1 FROM "T_T" WHERE "
			 "predicate=%lld AND object=%lld AND subject!=object "
			 "LIMIT 1", rspo, phash);
		num = ts_sql_count(symtab->c->db, cq);

		if (num) {
		    rasqal_variable rv;
		    rasqal_literal newvar;
		    raptor_uri *ru;
		    rasqal_literal spop;
		    char vname[TMP_SIZE];

		    handled = 1;

		    /* Rewrite the triple to use subproperty inference as
		     * we think there are subproperties:
		     *
		     * (?a, <foo:prop>, ?b) =>
		     *   (?a, _:tmp, ?b),
		     *   (_:tmp, <rdfs:subPropertyOf>, <foo:prop>)
		     */

		    newvar.type = RASQAL_LITERAL_VARIABLE;
		    snprintf(vname, TMP_SIZE, "_TV%d", symtab->tmpvar_cnt++);
		    rv.name = (unsigned char *)vname;
		    newvar.value.variable = &rv;

		    ru = raptor_new_uri((unsigned char *)RDFS_SUBPROPERTY);
		    spop.type = RASQAL_LITERAL_URI;
		    spop.value.uri = ru;

		    table_ob[symtab->hash_j_cnt] = idx;
		    snprintf(table, TMP_SIZE, "t%d", symtab->hash_j_cnt++);
		    ts_check_triple_slot(symtab, table, t->origin, "model", &sc[idx], idx);
		    ts_check_triple_slot(symtab, table, t->subject, "subject", &sc[idx], idx);
		    ts_check_triple_slot(symtab, table, &newvar, "predicate", &sc[idx], idx);
		    ts_check_triple_slot(symtab, table, t->object, "object", &sc[idx], idx);
		    table_ob[symtab->hash_j_cnt] = idx;
		    snprintf(table, TMP_SIZE, "t%d", symtab->hash_j_cnt++);
		    ts_check_triple_slot(symtab, table, t->origin, "model", &sc[idx], idx);
		    ts_check_triple_slot(symtab, table, &newvar, "subject", &sc[idx], idx);
		    ts_check_triple_slot(symtab, table, &spop, "predicate", &sc[idx], idx);
		    ts_check_triple_slot(symtab, table, t->predicate, "object", &sc[idx], idx);
		}
	    }
	}

	if (!handled) {
	    table_ob[symtab->hash_j_cnt] = idx;
	    snprintf(table, TMP_SIZE, "t%d", symtab->hash_j_cnt++);
	    ts_check_triple_slot(symtab, table, t->origin, "model", &sc[idx], idx);
	    ts_check_triple_slot(symtab, table, t->subject, "subject", &sc[idx], idx);
	    ts_check_triple_slot(symtab, table, t->predicate, "predicate", &sc[idx], idx);
	    ts_check_triple_slot(symtab, table, t->object, "object", &sc[idx], idx);
	}
    }

    return 0;
}

static void ts_check_triple_slot(ts_symtab *symtab, const char *table,
    rasqal_literal *l, const char *slot, GSList **sc, int block)
{
    ts_symbol *s;
    ts_hash hash;
    char tmp[TMP_SIZE];
#ifndef ENABLE_QUADS
    char mtbl[TMP_SIZE];
#endif /* ! ENABLE_QUADS */
    char *uri;
    int dt, lang;

    if (!strcmp("model", slot)) {
#ifndef ENABLE_QUADS
	slot = "mid";
	ts_symtab_need_model(symtab, table);
	sprintf(mtbl, "%sm", table);
	table = mtbl;
#endif /* ! ENABLE_QUADS */

	if (symtab->q->glimit) {
	    GString *gc = g_string_new("");
	    GSList *it;

	    for (it=symtab->q->glimit; it; it=it->next) {
		if (gc->len > 0) {
		    g_string_append(gc, " || ");
		}
		g_string_sprintfa(gc, "%s.%s=%lld", table, slot,
				  TS_ID(symtab->c, ts_hash_uri(it->data)));
	    }
	    *sc = g_slist_prepend(*sc, gc->str);
	    g_string_free(gc, FALSE);
	}
    }
    /* no constraints if theres no binding for this slot */
    if (!l) {
	/* this implements background graphs by requireing the model be 0 */
	if (symtab->c->feature[TS_FEATURE_STD_COMP]) {
	    snprintf(tmp, TMP_SIZE, "%s.%s=0", table, slot);
	    *sc = g_slist_prepend(*sc, strdup(tmp));
	}

	return;
    }

    switch (l->type) {
    case RASQAL_LITERAL_VARIABLE:
	s = ts_symtab_get(symtab, l->value.variable->name, FALSE, 0);
// printf("var(%s) appears in %d\n", l->value.variable->name, block);
	/* if the slot is NOT object */
	if (strcmp(slot, "object") && block == 0) {
	    s->bound_in_msp = 1;
	}

	if (s->hash_col) {
	    snprintf(tmp, TMP_SIZE, "%s.%s=%s", table, slot, ts_symbol_hashcol(s));
	    *sc = g_slist_prepend(*sc, strdup(tmp));
	    s->appears_in[block] = 1;
	} else {
	    snprintf(tmp, TMP_SIZE, "%s.%s", table, slot);
	    s->hash_col = strdup(tmp);
	    s->block = block;
	    s->appears_in[block] = 1;
	    if (ts_symbol_valtbl(s, block)) {
		snprintf(tmp, TMP_SIZE, "%s=%s."IDC, ts_symbol_hashcol(s),
			 ts_symbol_valtbl(s, block));
		*sc = g_slist_prepend(*sc, strdup(tmp));
	    }
	}
	break;

    case RASQAL_LITERAL_URI:
	uri = (char *)raptor_uri_to_string(l->value.uri);
	hash = TS_ID(symtab->c, ts_hash_uri(uri));
	free(uri);
	snprintf(tmp, TMP_SIZE, "%s.%s=%lld", table, slot, hash);
	*sc = g_slist_prepend(*sc, strdup(tmp));
	break;

    case RASQAL_LITERAL_STRING:
	uri = NULL;
	if (l->datatype) {
	    uri = (char *)raptor_uri_to_string(l->datatype);
	}
	if (uri) {
#ifdef ENABLE_QUADS
	    dt = ts_ensure_datatype(symtab->c, uri);
#else
	    dt = TS_ID(symtab->c, ts_hash_uri(uri));
#endif
	    free(uri);
	} else {
	    dt = 0;
	}
	if (l->language) {
#ifdef ENABLE_QUADS
	    lang = ts_ensure_language(symtab->c, l->language);
#else
	    lang = TS_ID(symtab->c, ts_hash_literal(l->language, 0, 0));
#endif
	} else {
	    lang = 0;
	}
	hash = TS_ID(symtab->c, ts_hash_literal((char *)(l->string), dt, lang));
	snprintf(tmp, TMP_SIZE, "%s.%s=%lld", table, slot, hash);
	*sc = g_slist_prepend(*sc, strdup(tmp));
	break;

    default:
	printf("lib3store: Erk, got a type of %d, do not know what to do\n", (int)l->type);
	break;
    }
}

static void ts_parse_constr_seq(ts_symtab *symtab, raptor_sequence *seq,
    GSList **sc, int block)
{
    int count;
    rasqal_expression *e;
    int i;
    int late;

    if (!seq) {
	return;
    }

    count = raptor_sequence_size(seq);
    for (i=0; i<count; i++) {
	e = raptor_sequence_get_at(seq, i);
	late = ts_expression_latebind(symtab, e);
#ifdef DEBUG
printf("con %d: ", i);
rasqal_expression_print(e, stdout);
printf("\n");
printf("LATE: %s\n", late ? "yes" : "no");
#endif
	if (!late) {
	    GString *s = g_string_new("");
	    ts_expression_descend(symtab, e, s, RASQAL_LITERAL_BOOLEAN, block);
	    /* sometimes, for reasons I dont understand this string is empty */
	    if (*(s->str)) {
		*sc = g_slist_prepend(*sc, strdup(s->str));
	    } else {
		printf("WARNING: empty expression:\n");
		rasqal_expression_print(e, stdout);
		printf("\n");
	    }
#ifdef DEBUG
printf("=> %s\n", s->str);
#endif
	    g_string_free(s, TRUE);
	} else {
	    symtab->q->globe = g_slist_prepend(symtab->q->globe, e);
	}
    }
}

static int regex_could_be_like(const char *re)
{
    const char *ptr;
    char last = '\0';

    for (ptr = re; *ptr; ptr++) {
	switch (*ptr) {
	case '[':
	case ']':
	case '(':
	case ')':
	case '+':
	    return 0;
	    break;
	case '*':
	    if (last != '.') {
		return 0;
	    }
	    break;
	}
	if (*ptr >= '0' && *ptr <= '9') {
	    if (last == '\\') {
		return 0;
	    }
	}
	last = *ptr;
    }

    return 1;
}

static void regex_to_like(const char *re, char *like)
{
    const char *ptr = re;
    char *optr = like;
    char last = '\0';

    if (*re != '^') {
	*(optr++) = '%';
    } else {
	ptr++;
    }
	
    while (*ptr) {
	if (*ptr == '.') {
	    if (*(ptr+1) == '*') {
		*(optr++) = '%';
		ptr+=2;
	    } else {
		*(optr++) = '?';
	    }
	    continue;
	}
	if (*ptr == '$') {
	    last = '$';
	    ptr++;
	    continue;
	}
	*(optr++) = *ptr;
	last = *ptr;
	ptr++;
    }
    if (last != '$') {
	*(optr++) = '%';
    }

    *optr = '\0';
}

static void ts_expression_descend(ts_symtab *st, rasqal_expression *e,
    GString *s, rasqal_literal_type prom, int block)
{
    ts_symbol *sym;
    char *column, *uri;
    char tmp[512];
    rasqal_expression *a1, *a2;

    if (!e) {
	g_string_append(s, "!ERROR!");

	return;
    }

    switch (e->op) {
    case RASQAL_EXPR_UNKNOWN:
	fprintf(stderr, "UNKNOWN operator encoutered\n");
	break;

    case RASQAL_EXPR_ORDER_COND_ASC:
	ts_expression_descend(st, e->arg1, s, TS_ORDER_EXPRESSION_ASC, block);
	break;

    case RASQAL_EXPR_ORDER_COND_DESC:
	ts_expression_descend(st, e->arg1, s, TS_ORDER_EXPRESSION_DESC, block);
	break;

    /* boolean infix boolean */
    case RASQAL_EXPR_AND:
    case RASQAL_EXPR_OR:
	g_string_append(s, "(");
	ts_expression_descend(st, e->arg1, s, RASQAL_LITERAL_BOOLEAN, block);
	g_string_append(s, ts_sql_operator(e->op));
	ts_expression_descend(st, e->arg2, s, RASQAL_LITERAL_BOOLEAN, block);
	g_string_append(s, ")");
	break;

    /* boolean infix values */
    case RASQAL_EXPR_EQ:
    case RASQAL_EXPR_NEQ:
    case RASQAL_EXPR_LT:
    case RASQAL_EXPR_GT:
    case RASQAL_EXPR_LE:
    case RASQAL_EXPR_GE:
    case RASQAL_EXPR_STR_EQ:
    case RASQAL_EXPR_STR_NEQ:
	prom = ts_expression_promoted_type(RASQAL_LITERAL_UNKNOWN, e->arg1);
	prom = ts_expression_promoted_type(prom, e->arg2);

	/* If were using optimised evaluation then we needs to make sure we
         * dont try to evaluate const = ?var */
	if (st->c->opt_lvl > 0 && e->arg2->op == RASQAL_EXPR_LITERAL && 
	    e->arg2->literal->type == RASQAL_LITERAL_VARIABLE &&
	    !(e->arg1->literal->type == RASQAL_LITERAL_VARIABLE &&
	      e->arg1->op != RASQAL_EXPR_LITERAL)) {
	    a1 = e->arg2;
	    a2 = e->arg1;
	} else {
	    a1 = e->arg1;
	    a2 = e->arg2;
	}
	g_string_append(s, "(");
	ts_expression_descend(st, a1, s, prom, block);
	g_string_append(s, ts_sql_operator(e->op));
	ts_expression_descend(st, a2, s, prom, block);
	if (e->op == RASQAL_EXPR_NEQ) {
	    add_known_type_constraint(s, e->arg1, st, block);
	    add_known_type_constraint(s, e->arg2, st, block);
	}
	g_string_append(s, ")");
	break;

    /* regex */
    case RASQAL_EXPR_STR_MATCH:
    case RASQAL_EXPR_STR_NMATCH:
	g_string_append(s, "(");
	if (st->c->opt_lvl > 0 &&
	    regex_could_be_like((char *)e->literal->string)) {
	    char *like = malloc(strlen((char *)e->literal->string) + 4);

	    regex_to_like((char *)e->literal->string, like);
	    ts_expression_descend(st, e->arg1, s, RASQAL_LITERAL_STRING, block);
	    g_string_append(s, " LIKE ");
	    g_string_sprintfa(s, "'%s')", like);
	    free(like);
	} else {
	    ts_expression_descend(st, e->arg1, s, RASQAL_LITERAL_STRING, block);
	    g_string_append(s, ts_sql_operator(e->op));
	    g_string_sprintfa(s, "'%s')", e->literal->string);
	}
	break;

    case RASQAL_EXPR_TILDE:
	g_string_append(s, "~(");
	ts_expression_descend(st, e->arg1, s, RASQAL_LITERAL_UNKNOWN, block);
	g_string_append(s, ")");
	break;
	
    case RASQAL_EXPR_BANG:
	g_string_append(s, "NOT (");
	ts_expression_descend(st, e->arg1, s, RASQAL_LITERAL_BOOLEAN, block);
	g_string_append(s, ")");
	break;
	
    case RASQAL_EXPR_PLUS:
    case RASQAL_EXPR_MINUS:
    case RASQAL_EXPR_STAR:
    case RASQAL_EXPR_REM:
	prom = ts_expression_promoted_type(RASQAL_LITERAL_UNKNOWN, e->arg1);
	prom = ts_expression_promoted_type(prom, e->arg2);
	if (prom == RASQAL_LITERAL_UNKNOWN) prom = RASQAL_LITERAL_FLOATING;
	g_string_append(s, "(");
	ts_expression_descend(st, e->arg1, s, prom, block);
	g_string_append(s, ts_sql_operator(e->op));
	ts_expression_descend(st, e->arg2, s, prom, block);
	g_string_append(s, ")");
	break;

    case RASQAL_EXPR_SLASH:
	g_string_append(s, "(");
	ts_expression_descend(st, e->arg1, s, RASQAL_LITERAL_FLOATING, block);
	g_string_append(s, ts_sql_operator(e->op));
	ts_expression_descend(st, e->arg2, s, RASQAL_LITERAL_FLOATING, block);
	g_string_append(s, ")");
	break;

    case RASQAL_EXPR_UMINUS:
	g_string_append(s, "-(");
	ts_expression_descend(st, e->arg1, s, prom, block);
	g_string_append(s, ")");
	break;

    case RASQAL_EXPR_LITERAL:
	switch (e->literal->type) {
	case RASQAL_LITERAL_UNKNOWN:
	    g_string_sprintfa(s, "!UNKNOWN!");
	    break;

	case RASQAL_LITERAL_BLANK:
	case RASQAL_LITERAL_URI: {
	    char *u = (char *)raptor_uri_to_string(e->literal->value.uri);
	    g_string_sprintfa(s, "\"%s\"",u);
	    free(u);
	    break;
	    }

	case RASQAL_LITERAL_PATTERN:
	case RASQAL_LITERAL_STRING:
	    uri = NULL;
	    if (e->literal->datatype)
		uri = (char *)raptor_uri_to_string(e->literal->datatype);
	    if (uri && !strcmp(uri, XSD_DATETIME)) {
		snprintf(tmp, 512, "'%s'", e->literal->string);
		g_string_append(s, ts_normalise_date(st->c, st->c->datetime_dt,
			    tmp, tmp));
	    } else {
		g_string_sprintfa(s, "\"%s\"", e->literal->string);
	    }
	    free(uri);
	    break;

	case RASQAL_LITERAL_DATETIME:
	    snprintf(tmp, 512, "CAST('%s' AS datetime)", e->literal->string);
	    g_string_append(s, ts_normalise_date(st->c, st->c->datetime_dt,
			tmp, tmp));
	    break;

	case RASQAL_LITERAL_INTEGER:
	    g_string_sprintfa(s, "%i", e->literal->value.integer);
	    break;

	case RASQAL_LITERAL_FLOATING:
	case RASQAL_LITERAL_FLOAT:
	case RASQAL_LITERAL_DECIMAL:
	    g_string_sprintfa(s, "%0.16g", e->literal->value.floating);
	    break;

	case RASQAL_LITERAL_VARIABLE: {
	    ts_symbol *lsym = ts_symtab_get(st, e->literal->value.variable->name, TRUE, block);
	    char *valtbl = (char *)ts_symbol_valtbl(lsym, block);
	    char *vtable;
	    column = "val";

	    switch (prom) {
	    case RASQAL_LITERAL_FLOAT:
	    case RASQAL_LITERAL_DOUBLE:
		vtable = (char *)ts_symbol_valtbl_type(lsym, block, RASQAL_LITERAL_FLOAT);
		break;
	    case RASQAL_LITERAL_INTEGER:
		vtable = (char *)ts_symbol_valtbl_type(lsym, block, RASQAL_LITERAL_INTEGER);
		break;
	    case RASQAL_LITERAL_DATETIME:
		vtable = (char *)ts_symbol_valtbl_type(lsym, block, RASQAL_LITERAL_DATETIME);
		break;
	    case RASQAL_LITERAL_BOOLEAN:
		column = "ebv";
		vtable = valtbl;
		break;
	    default:
		column = "lexical";
		vtable = valtbl;
		break;
	    }

	    if (prom == TS_ORDER_EXPRESSION_ASC || prom == TS_ORDER_EXPRESSION_DESC) {
		char *dir = (prom == TS_ORDER_EXPRESSION_ASC) ? "ASC" : "DESC";
		const char *vftable = ts_symbol_valtbl_type(lsym, TS_ORDER_BLOCK, RASQAL_LITERAL_FLOATING);
		const char *fcolumn = "val";

		ts_symbol_valtbl_reqd(lsym, TS_ORDER_BLOCK);
		g_string_sprintfa(s, "((%s.hash >= 0)*2 - (%s.hash < 0 && LEFT(%s.lexical, 2)='_:')) %s, %s.%s %s, ", valtbl, valtbl, valtbl, dir, vftable, fcolumn, dir);
	    }

	    if (prom == RASQAL_LITERAL_URI) {
		char tmp[TMP_SIZE];
		snprintf(tmp, TMP_SIZE-1, "%s.hash<0", valtbl);
		if (st->c->opt_lvl > 0) {
		    g_string_sprintfa(s, "%s && %s.%s", tmp, vtable, column);
		} else {
		    g_string_sprintfa(s, "if (%s, %s.%s, NULL)", tmp,
				      vtable, column);
		}
	    } else if (prom == RASQAL_LITERAL_STRING) {
		char tmp[TMP_SIZE];
		snprintf(tmp, TMP_SIZE-1, "%s.hash>=0", ts_symbol_valtbl(lsym, block));
		if (st->c->opt_lvl > 0) {
		    g_string_sprintfa(s, "%s && %s.%s", tmp, vtable, column);
		} else {
		    g_string_sprintfa(s, "if (%s, %s.%s, NULL)", tmp,
				      vtable, column);
		}
	    } else if (prom == TS_ORDER_EXPRESSION_ASC) {
		g_string_sprintfa(s, "%s.%s", vtable, column);
		g_string_append(s, " ASC");
		g_string_sprintfa(s, ", %s.datatype ASC", valtbl);
	    } else if (prom == TS_ORDER_EXPRESSION_DESC) {
		g_string_sprintfa(s, "%s.%s", vtable, column);
		g_string_append(s, " DESC");
		g_string_sprintfa(s, ", %s.datatype DESC", valtbl);
	    } else {
		g_string_sprintfa(s, "%s.%s", vtable, column);
	    }
	    }
	    break;

	case RASQAL_LITERAL_BOOLEAN:
	    g_string_sprintfa(s, "(%d!=0)", e->literal->value.integer);
	    break;

	case RASQAL_LITERAL_QNAME:
	    g_string_sprintfa(s, "!ERROR!");
	    break;
	    
#if 0
	default:
	    g_string_sprintfa(s, "!UNKNOWN!");
	    break;
#endif
	}
	break;

    case RASQAL_EXPR_FUNCTION: {
	char *typeuri = (char *)raptor_uri_as_string(e->name);
	rasqal_expression *arg0 = raptor_sequence_get_at(e->args, 0);

	/* look to see if its probably a cast, if its an XSD type... */
	if (arg0 && !strncmp(XSD_NAMESPACE, typeuri, 33)) {

	    if (arg0->op == RASQAL_EXPR_LITERAL &&
		arg0->literal->type == RASQAL_LITERAL_VARIABLE) {
		char *column;
		char *vtable;

		sym = ts_symtab_get(st, arg0->literal->value.variable->name, TRUE, block);
		if (!strcmp(typeuri, XSD_FLOAT) ||
		    !strcmp(typeuri, XSD_DOUBLE)) {
		    vtable = (char *)ts_symbol_valtbl_type(sym, block,
						   RASQAL_LITERAL_FLOAT);
		    column = "val";
		} else if (!strcmp(typeuri, XSD_INTEGER) ||
			   !strcmp(typeuri, XSD_BOOLEAN)) {
		    vtable = (char *)ts_symbol_valtbl_type(sym, block,
						   RASQAL_LITERAL_INTEGER);
		    column = "val";
		} else if (!strcmp(typeuri, XSD_DATE) ||
			   !strcmp(typeuri, XSD_TIME) ||
			   !strcmp(typeuri, XSD_DATETIME)) {
		    vtable = (char *)ts_symbol_valtbl_type(sym, block,
						   RASQAL_LITERAL_DATETIME);
		    column = "val";
		} else {
		    vtable = (char *)ts_symbol_valtbl(sym, block);
		    column = "lexical";
		}
		g_string_sprintfa(s, "%s.%s", vtable, column);
	    } else if (arg0->op == RASQAL_EXPR_LITERAL &&
		       arg0->literal->type == RASQAL_LITERAL_STRING) {
		char *val = (char *)arg0->literal->string;

		if (!strcmp(typeuri, XSD_FLOAT) ||
		    !strcmp(typeuri, XSD_DOUBLE)) {
		    g_string_sprintfa(s, "(%g+0.0)", atof(val));
		} else if (!strcmp(typeuri, XSD_INTEGER)) {
		    g_string_sprintfa(s, "(%lld)", atoll(val));
		} else if (!strcmp(typeuri, XSD_DATE) ||
			   !strcmp(typeuri, XSD_TIME) ||
			   !strcmp(typeuri, XSD_DATETIME)) {
		    char tmp[512];
		    char *dt;
		    int len = strlen(val);

		    if (len > 250) {
			len = 250;
		    }

		    mysql_real_escape_string(st->c->db, tmp, val, len);
		    dt = ts_normalise_date(st->c, st->c->datetime_dt, tmp, tmp);
		    g_string_sprintfa(s, "%s", dt);
		} else {
		    g_string_sprintfa(s, "'%s'", val);
		}
	    } else {
		ts_expression_descend(st, e->arg1, s, prom, block);
	    }
	/* It's a function of some kind, or an unknown cast */
	} else {
	    g_string_sprintfa(s, "func /* %s */", raptor_uri_to_string(e->name));
	}

	break;
    }

    case RASQAL_EXPR_BOUND:
	ts_expression_descend(st, e->arg1, s, RASQAL_LITERAL_UNKNOWN, block);
	g_string_append(s, " IS NOT NULL");
	break;

    case RASQAL_EXPR_STR:
	if (e->arg1->op == RASQAL_EXPR_LITERAL &&
            e->arg1->literal->type == RASQAL_LITERAL_VARIABLE) {
	    sym = ts_symtab_get(st, e->arg1->literal->value.variable->name,
				TRUE, block);
	    g_string_sprintfa(s, "if (%s.hash<0 && LEFT(%s.lexical, 2)='_:', "
				 "'', " "%s.lexical)",
			    ts_symbol_valtbl(sym, block),
			    ts_symbol_valtbl(sym, block),
			    ts_symbol_valtbl(sym, block));
	} else {
	    ts_expression_descend(st, e->arg1, s, RASQAL_LITERAL_UNKNOWN, block);
	}
	break;

    case RASQAL_EXPR_LANG:
	sym = ts_symtab_get(st, e->arg1->literal->value.variable->name, TRUE, block);
	ts_symbol_need_lang(st, sym);
	g_string_sprintfa(s, "%s.lang", sym->lang_col);
	break;

    case RASQAL_EXPR_DATATYPE:
	sym = ts_symtab_get(st, e->arg1->literal->value.variable->name, TRUE, block);
	ts_symbol_need_dt(st, sym);
	g_string_sprintfa(s, "%s.uri", sym->dt_col);
	break;

    case RASQAL_EXPR_ISURI:
	sym = ts_symtab_get(st, e->arg1->literal->value.variable->name, TRUE, block);
	g_string_sprintfa(s, "(%s.hash<0 && LEFT(%s.lexical, 2)!='_:')",
			  ts_symbol_valtbl(sym, block),
			  ts_symbol_valtbl(sym, block));
	break;

    case RASQAL_EXPR_ISBLANK:
	sym = ts_symtab_get(st, e->arg1->literal->value.variable->name, TRUE, block);
	g_string_sprintfa(s, "%s.hash<0 && LEFT(%s.lexical, 2)='_:'",
	                  ts_symbol_valtbl(sym, block),
			  ts_symbol_valtbl(sym, block));
	break;

    case RASQAL_EXPR_ISLITERAL:
#ifdef ENABLE_QUADS
	sym = ts_symtab_get(st, e->arg1->literal->value.variable->name, FALSE, block);
	g_string_sprintfa(s, "%s >= 0", sym->hash_col);
#else
	sym = ts_symtab_get(st, e->arg1->literal->value.variable->name, TRUE, block);
	g_string_sprintfa(s, "%s.hash >= 0", ts_symbol_valtbl(sym, block));
#endif
	break;

    case RASQAL_EXPR_CAST: {
	char *typeuri = (char *)raptor_uri_as_string(e->name);

	if (e->arg1->op == RASQAL_EXPR_LITERAL &&
	    e->arg1->literal->type == RASQAL_LITERAL_VARIABLE) {
	    char *column;
	    char *vtable;

	    sym = ts_symtab_get(st, e->arg1->literal->value.variable->name, TRUE, block);
	    if (!strcmp(typeuri, XSD_FLOAT) ||
		!strcmp(typeuri, XSD_DOUBLE)) {
		vtable = (char *)ts_symbol_valtbl_type(sym, block,
					       RASQAL_LITERAL_FLOAT);
		column = "val";
	    } else if (!strcmp(typeuri, XSD_INTEGER) ||
		       !strcmp(typeuri, XSD_BOOLEAN)) {
		vtable = (char *)ts_symbol_valtbl_type(sym, block,
					       RASQAL_LITERAL_INTEGER);
		column = "val";
	    } else if (!strcmp(typeuri, XSD_DATE) ||
		       !strcmp(typeuri, XSD_TIME) ||
		       !strcmp(typeuri, XSD_DATETIME)) {
		vtable = (char *)ts_symbol_valtbl_type(sym, block,
					       RASQAL_LITERAL_DATETIME);
		column = "val";
	    } else {
		vtable = (char *)ts_symbol_valtbl(sym, block);
		column = "lexical";
	    }
	    g_string_sprintfa(s, "%s.%s", vtable, column);
	} else {
	    ts_expression_descend(st, e->arg1, s, prom, block);
	}

	break;
    }

#if 0
    default:
	{
	    char enumval[64];
	    sprintf(enumval, "%d", e->op);
	    g_string_append(s, "???");
	    g_string_append(s, enumval);
	    g_string_append(s, "???");
	    break;
	}
#endif
    }
}

static void add_known_type_constraint(GString *s, rasqal_expression *e,
					ts_symtab *st, int block)
{
    if (e->op == RASQAL_EXPR_LITERAL &&
	e->literal->type == RASQAL_LITERAL_VARIABLE) {
	ts_symbol *sym = ts_symtab_get(st, e->literal->value.variable->name,
				       TRUE, block);
	g_string_sprintfa(s, " && %s.datatype IN (0,%d,%d,%d,%d,%d)",
	    ts_symbol_valtbl(sym, block), st->c->int_dt, st->c->string_dt,
	    st->c->float_dt, st->c->double_dt, st->c->datetime_dt);
    }
}

static int ts_expression_latebind(ts_symtab *st, rasqal_expression *e)
{
    ts_symbol *sym;

    if (!e) {
	return 0;
    }

    switch (e->op) {
    case RASQAL_EXPR_UNKNOWN:
	return 0;

    case RASQAL_EXPR_ORDER_COND_ASC:
    case RASQAL_EXPR_ORDER_COND_DESC:
	return 0;

    /* 2 arg things */
    case RASQAL_EXPR_AND:
    case RASQAL_EXPR_OR:
    case RASQAL_EXPR_EQ:
    case RASQAL_EXPR_NEQ:
    case RASQAL_EXPR_LT:
    case RASQAL_EXPR_GT:
    case RASQAL_EXPR_LE:
    case RASQAL_EXPR_GE:
    case RASQAL_EXPR_STR_EQ:
    case RASQAL_EXPR_STR_NEQ:
    case RASQAL_EXPR_STR_MATCH:
    case RASQAL_EXPR_STR_NMATCH:
    case RASQAL_EXPR_PLUS:
    case RASQAL_EXPR_MINUS:
    case RASQAL_EXPR_STAR:
    case RASQAL_EXPR_REM:
    case RASQAL_EXPR_SLASH:
	return ts_expression_latebind(st, e->arg1) +
	       ts_expression_latebind(st, e->arg2);

    /* 1 arg things */
    case RASQAL_EXPR_TILDE:
    case RASQAL_EXPR_BANG:
    case RASQAL_EXPR_UMINUS:
    case RASQAL_EXPR_BOUND:
    case RASQAL_EXPR_STR:
    case RASQAL_EXPR_LANG:
    case RASQAL_EXPR_DATATYPE:
    case RASQAL_EXPR_ISURI:
    case RASQAL_EXPR_ISBLANK:
    case RASQAL_EXPR_ISLITERAL:
    case RASQAL_EXPR_CAST:
	return ts_expression_latebind(st, e->arg1);
	
    case RASQAL_EXPR_LITERAL:
	switch (e->literal->type) {
	    /* constants */
	case RASQAL_LITERAL_UNKNOWN:
	case RASQAL_LITERAL_BLANK:
	case RASQAL_LITERAL_URI:
	case RASQAL_LITERAL_PATTERN:
	case RASQAL_LITERAL_STRING:
	case RASQAL_LITERAL_DATETIME:
	case RASQAL_LITERAL_INTEGER:
	case RASQAL_LITERAL_FLOATING:
	case RASQAL_LITERAL_FLOAT:
	case RASQAL_LITERAL_DECIMAL:
	case RASQAL_LITERAL_BOOLEAN:
	case RASQAL_LITERAL_QNAME:
	    return 0;

	case RASQAL_LITERAL_VARIABLE:
	    sym = ts_symtab_get(st, e->literal->value.variable->name, FALSE, 0);
	    if (sym->hash_col) {
		return 0;
	    }
	    return 1;
	    
#if 0
	default:
	    g_string_sprintfa(s, "!UNKNOWN!");
	    break;
#endif
	}

    case RASQAL_EXPR_FUNCTION: {
	int ret = 0;
	int pos = 0;
	rasqal_expression *arg;

	while ((arg = raptor_sequence_get_at(e->args, pos++))) {
	    ret += ts_expression_latebind(st, arg);
	}

	return ret;
    }

#if 0
    default:
	{
	    char enumval[64];
	    sprintf(enumval, "%d", e->op);
	    g_string_append(s, "???");
	    g_string_append(s, enumval);
	    g_string_append(s, "???");
	    break;
	}
#endif
    }

    return 0;
}

static const char *ts_sql_operator(rasqal_op op)
{
    switch (op) {
    case RASQAL_EXPR_AND:
	return "&&";
    case RASQAL_EXPR_OR:
	return "||";
    case RASQAL_EXPR_EQ:
	return "=";
    case RASQAL_EXPR_NEQ:
	return "!=";
    case RASQAL_EXPR_LT:
	return "<";
    case RASQAL_EXPR_GT:
	return ">";
    case RASQAL_EXPR_LE:
	return "<=";
    case RASQAL_EXPR_GE:
	return ">=";
    case RASQAL_EXPR_UMINUS:
	return "-";
    case RASQAL_EXPR_PLUS:
	return "+";
    case RASQAL_EXPR_MINUS:
	return "-";
    case RASQAL_EXPR_STAR:
	return "*";
    case RASQAL_EXPR_SLASH:
	return "/";
    case RASQAL_EXPR_REM:
	return " % ";
    case RASQAL_EXPR_STR_EQ:
	return "=";
    case RASQAL_EXPR_STR_NEQ:
	return "!=";
    case RASQAL_EXPR_STR_MATCH:
	return " REGEXP ";
    case RASQAL_EXPR_STR_NMATCH:
	return " !REGEXP ";
    case RASQAL_EXPR_TILDE:
	return "~";
    case RASQAL_EXPR_BANG:
	return "!";
    default:
	return " ?OP? ";
    }
}

static rasqal_literal_type ts_expression_promoted_type(rasqal_literal_type t,
    rasqal_expression *e)
{
    if (e->op == RASQAL_EXPR_LITERAL) {
	rasqal_literal_type u = e->literal->type;

	if ((u == RASQAL_LITERAL_UNKNOWN || u == RASQAL_LITERAL_STRING) &&
            e->literal->datatype) {
	    char *dt = (char *)raptor_uri_to_string(e->literal->datatype);

	    if (!strcmp(dt, XSD_DATETIME)) {
		u = RASQAL_LITERAL_DATETIME;
	    } else if (!strcmp(dt, XSD_STRING)) {
		u = RASQAL_LITERAL_STRING;
	    } else if (!strcmp(dt, XSD_INTEGER)) {
		u = RASQAL_LITERAL_INTEGER;
	    } else if (!strcmp(dt, XSD_DOUBLE)) {
		u = RASQAL_LITERAL_DOUBLE;
	    } else if (!strcmp(dt, XSD_FLOAT)) {
		u = RASQAL_LITERAL_FLOAT;
	    }
	}

	if (t == RASQAL_LITERAL_UNKNOWN) {
	    return u;
	}

	if (t == RASQAL_LITERAL_BOOLEAN) {
	    t = RASQAL_LITERAL_INTEGER;
	}
	if (u == RASQAL_LITERAL_BOOLEAN) {
	    u = RASQAL_LITERAL_INTEGER;
	}

	if (t == RASQAL_LITERAL_DOUBLE || u == RASQAL_LITERAL_DOUBLE) {
	    return RASQAL_LITERAL_DOUBLE;
	}
	if (t == RASQAL_LITERAL_FLOAT || u == RASQAL_LITERAL_FLOAT) {
	    return RASQAL_LITERAL_FLOAT;
	}
	if (t == RASQAL_LITERAL_INTEGER || u == RASQAL_LITERAL_INTEGER) {
	    return RASQAL_LITERAL_INTEGER;
	}
	if (t == RASQAL_LITERAL_DATETIME || u == RASQAL_LITERAL_DATETIME) {
	    return RASQAL_LITERAL_DATETIME;
	}
	if (t == RASQAL_LITERAL_STRING || u == RASQAL_LITERAL_STRING) {
	    return RASQAL_LITERAL_STRING;
	}
	if (t == RASQAL_LITERAL_URI || u == RASQAL_LITERAL_URI) {
	    return RASQAL_LITERAL_URI;
	}
    }
    if (e->op == RASQAL_EXPR_FUNCTION) {
	char *typeuri = (char *)raptor_uri_as_string(e->name);

	/* if its an XSD type */
	if (!strncmp(XSD_NAMESPACE, typeuri, 33)) {
	    if (!strcmp(typeuri, XSD_FLOAT) ||
		!strcmp(typeuri, XSD_DOUBLE)) {
		return RASQAL_LITERAL_FLOATING;
	    } else if (!strcmp(typeuri, XSD_INTEGER)) {
		return RASQAL_LITERAL_INTEGER;
	    } else if (!strcmp(typeuri, XSD_DATE) ||
		       !strcmp(typeuri, XSD_TIME) ||
		       !strcmp(typeuri, XSD_DATETIME)) {
		return RASQAL_LITERAL_DATETIME;
	    } else {
		return RASQAL_LITERAL_STRING;
	    }
	}
    }
    if (ts_boolean_op(e)) {
	return RASQAL_LITERAL_INTEGER;
    }
    if (e->op == RASQAL_EXPR_SLASH) {
	return RASQAL_LITERAL_FLOATING;
    }
    if (ts_binary_numerical_op(e)) {
	rasqal_literal_type nt = ts_expression_promoted_type(t, e->arg1);
	return ts_expression_promoted_type(nt, e->arg2);
    }

    return t;
}

static int ts_boolean_op(rasqal_expression *e)
{
    switch (e->op) {
    case RASQAL_EXPR_UNKNOWN:
    case RASQAL_EXPR_ORDER_COND_ASC:
    case RASQAL_EXPR_ORDER_COND_DESC:
	return FALSE;

    case RASQAL_EXPR_AND:
    case RASQAL_EXPR_OR:
    case RASQAL_EXPR_EQ:
    case RASQAL_EXPR_NEQ:
    case RASQAL_EXPR_LT:
    case RASQAL_EXPR_GT:
    case RASQAL_EXPR_LE:
    case RASQAL_EXPR_GE:
	return TRUE;

    case RASQAL_EXPR_UMINUS:
    case RASQAL_EXPR_PLUS:
    case RASQAL_EXPR_MINUS:
    case RASQAL_EXPR_STAR:
    case RASQAL_EXPR_SLASH:
    case RASQAL_EXPR_REM:
	return FALSE;

    case RASQAL_EXPR_STR_EQ:
    case RASQAL_EXPR_STR_NEQ:
    case RASQAL_EXPR_STR_MATCH:
    case RASQAL_EXPR_STR_NMATCH:
	return TRUE;

    case RASQAL_EXPR_TILDE:
    case RASQAL_EXPR_BANG:
    case RASQAL_EXPR_LITERAL:
	return FALSE;

    case RASQAL_EXPR_FUNCTION:
    case RASQAL_EXPR_BOUND:
    case RASQAL_EXPR_ISURI:
    case RASQAL_EXPR_ISBLANK:
    case RASQAL_EXPR_ISLITERAL:
	return TRUE;

    case RASQAL_EXPR_STR:
    case RASQAL_EXPR_LANG:
    case RASQAL_EXPR_DATATYPE:
    case RASQAL_EXPR_CAST:
	return FALSE;
    }

    return FALSE;
}

static int ts_binary_numerical_op(rasqal_expression *e)
{
    switch (e->op) {
    case RASQAL_EXPR_UNKNOWN:
    case RASQAL_EXPR_ORDER_COND_ASC:
    case RASQAL_EXPR_ORDER_COND_DESC:
    case RASQAL_EXPR_AND:
    case RASQAL_EXPR_OR:
    case RASQAL_EXPR_EQ:
    case RASQAL_EXPR_NEQ:
    case RASQAL_EXPR_LT:
    case RASQAL_EXPR_GT:
    case RASQAL_EXPR_LE:
    case RASQAL_EXPR_GE:
    case RASQAL_EXPR_UMINUS:
	return FALSE;

    case RASQAL_EXPR_PLUS:
    case RASQAL_EXPR_MINUS:
    case RASQAL_EXPR_STAR:
    case RASQAL_EXPR_SLASH:
	return TRUE;

    case RASQAL_EXPR_REM:
    case RASQAL_EXPR_STR_EQ:
    case RASQAL_EXPR_STR_NEQ:
    case RASQAL_EXPR_STR_MATCH:
    case RASQAL_EXPR_STR_NMATCH:
    case RASQAL_EXPR_TILDE:
    case RASQAL_EXPR_BANG:
    case RASQAL_EXPR_LITERAL:
    case RASQAL_EXPR_CAST:
	return FALSE;

    case RASQAL_EXPR_FUNCTION:
    case RASQAL_EXPR_BOUND:
    case RASQAL_EXPR_ISURI:
    case RASQAL_EXPR_ISBLANK:
    case RASQAL_EXPR_ISLITERAL:
    case RASQAL_EXPR_STR:
    case RASQAL_EXPR_LANG:
    case RASQAL_EXPR_DATATYPE:
	return FALSE;
    }

    return FALSE;
}

/* vi:set ts=8 sts=4 sw=4: */
