/*
 *  Copyright (C) 2005 Steve Harris
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  $Id: symtab.c,v 1.13 2005/11/10 16:51:13 theno23 Exp $
 */

#include <stdlib.h>
#include <string.h>

#include "tstore.h"
#include "config.h"

ts_symbol *ts_symbol_new(const char *sym, const char *hash_col)
{
    ts_symbol *s = calloc(1, sizeof(ts_symbol));

    if (sym) s->sym = strdup(sym);
    if (hash_col) s->hash_col = strdup(hash_col);
    s->block = -1;

    return s;
}

void ts_symbol_free(ts_symbol *s)
{
    int i;

    if (s) {
	free(s->sym);
	free(s->hash_col);
	for (i=0; i<TS_MAX_BLOCKS; i++) {
	    if (s->val_tbl[i]) free(s->val_tbl[i]);
	}
	free(s->tt_col);
    }
    free(s);
}

const char *ts_symbol_hashcol(ts_symbol *s)
{
    if (s->tt_col) return s->tt_col;

    return s->hash_col;
}

const char *ts_symbol_valtbl(ts_symbol *s, int block)
{
    if (block < 0 || block >= TS_MAX_BLOCKS) {
	return "!out of range!";
    }

    if (block == TS_ORDER_BLOCK && s->val_tbl[0]) {
	return s->val_tbl[0];
    }

    return s->val_tbl[block];
}

const char *ts_symbol_valtbl_type(ts_symbol *s, int block, 
				  rasqal_literal_type type)
{
    char *tag;
    char *table;
    int idx;

    if (block < 0 || block >= TS_MAX_BLOCKS) {
	return "!out of range!";
    }

    switch (type) {
    case RASQAL_LITERAL_INTEGER:
	table = T_I;
	tag = "i";
	idx = 0;
	break;
    case RASQAL_LITERAL_FLOATING:
	table = T_F;
	tag = "f";
	idx = 1;
	break;
    case RASQAL_LITERAL_DATETIME:
	table = T_A;
	tag = "d";
	idx = 2;
	break;
    default:
	return "!TYPE ERROR in ts_symbol_valtbl_type!";
	break;
    }
	
    if (!s->val_tbl_type[idx][block]) {
	s->val_tbl_type[idx][block] = g_strdup_printf("v%s%d", tag, s->st->val_j_cnt++);
    }

    return s->val_tbl_type[idx][block];
}

const char *ts_symbol_valtbl_reqd(ts_symbol *s, int block)
{
    if (block < 0 || block >= TS_MAX_BLOCKS) {
	return "!out of range!";
    }

    if (!s->val_tbl[block]) {
	s->val_tbl[block] = g_strdup_printf("v%d", s->st->val_j_cnt++);
    }

    return s->val_tbl[block];
}

int ts_symbol_appears_in(ts_symbol *s, int block)
{
    if (block < 0 || block > TS_MAX_BLOCKS) {
	return 0;
    }

    return s->appears_in[block];
}

void ts_symbol_need_dt(ts_symtab *st, ts_symbol *s)
{
    if (s->dt_col == NULL) {
	s->dt_col = g_strdup_printf("d%d", st->dt_j_cnt++);
    }
}

void ts_symbol_need_lang(ts_symtab *st, ts_symbol *s)
{
    if (s->lang_col == NULL) {
	s->lang_col = g_strdup_printf("l%d", st->lang_j_cnt++);
    }
}

#ifndef ENABLE_QUADS
void ts_symtab_need_model(ts_symtab *symtab, const char *tbl)
{
    int tbl_num = atoi(tbl+1); // yuk!

    symtab->model_joins[tbl_num] = 1;
}
#endif /* ! ENABLE_QUADS */

/* look up a symbol in the symbol table, and return it
 *
 * if val_needed is true then make sure it has a table name for the symbol
 * value to be joined to
 */
ts_symbol *ts_symtab_get_intl(ts_symtab *symtab, const unsigned char *name,
	int val_needed, int block)
{
    ts_symbol *s = g_hash_table_lookup(symtab->hash, name);

    if (!s) {
	s = ts_symbol_new((char *)name, (char *)NULL);
	g_hash_table_insert(symtab->hash, (char *)name, s);
	s->st = symtab;
    }

    if (val_needed && !s->val_tbl[block]) {
	s->val_tbl[block] = g_strdup_printf("v%d", symtab->val_j_cnt++);
    }

    return s;
}

void print_entry(gpointer key, gpointer val, gpointer data)
{
    int i;

    ts_symbol *s = (ts_symbol *)val;
    fprintf((FILE *)data, "   %s - h:%s", (char *)key, s->hash_col);
    for (i=0; i<TS_MAX_BLOCKS; i++) {
	if (s->val_tbl[i]) {
	    fprintf((FILE *)data, " v%d:%s", i, s->val_tbl[i]);
	}
    }
    fprintf((FILE *)data, "\n");
}

static void value_destroy(gpointer data)
{
    ts_symbol_free(data);
}

void ts_symtab_dump(ts_symtab *symtab, FILE *out)
{
    fprintf(out, "Symtab:\n");
    g_hash_table_foreach(symtab->hash, print_entry, out);
}

ts_symtab *ts_symtab_new(ts_connection *c, ts_query *q)
{
    ts_symtab *ts = calloc(1, sizeof(ts_symtab));
    ts->hash = g_hash_table_new_full(g_str_hash, g_str_equal, NULL,
				     value_destroy);
    ts->c = c;
    ts->q = q;

    return ts;
}

void ts_symtab_free(ts_symtab *symtab)
{
    g_hash_table_destroy(symtab->hash);
    free(symtab);
}

/* vi:set ts=8 sts=4 sw=4: */
