/*
 *  Copyright (C) 2005 Steve Harris
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  $Id: ts-import.c,v 1.22 2005/11/16 18:46:51 theno23 Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libgen.h>
#include <rasqal.h>

#include "tstore.h"
#include "bulk-import.h"
#include "timing.h"
#include "cmdline-parser.h"
#include "sys/time.h"

int main(int argc, char*argv[])
{
    ts_connection *con = NULL;
    int help = 0;
    char *optstring = "-" TSP_OPTSTR "m:f:bvaBlL";
    char *model = NULL;
    char *cmodel = NULL;
    char *uri = NULL;
    int c, opt_index = 0;
    long long erased = 0; /* count of erased triples */
    int verbosity = 0;
    int add = 0;
    int bulk = 0;
    int count = 0;
    int done = 0;
    int lock = 1;
    ts_syntax format = TS_AUTO;
    struct timeval then, now;
    double diff;
    static struct option long_options[] = {
        TSP_LONGOPTS,
        { "verbose", 0, 0, 'v' },
        { "model", 1, 0, 'm' },
        { "format", 1, 0, 'f' },
        { "background", 0, 0, 'b' },
        { "bulk", 0, 0, 'B' },
        { "add", 0, 0, 'a' },
        { "lock", 0, 0, 'l' },
        { "no-lock", 0, 0, 'L' },
	{ 0, 0, 0, 0 }
    };

    gettimeofday(&then, 0);

    ts_init();

    while ((c = getopt_long (argc, argv, optstring, long_options, &opt_index)) != -1) {
        if (!tsp_handle_option(c, optarg)) {
            switch (c) {
            case 'h':
                help++;
                break;
	    case 'v':
		verbosity++;
	    case 'm':
		model = optarg;
		break;
	    case 'b':
		model = "background";
		break;
	    case 'B':
		bulk = 1;
		break;
	    case 'a':
		add = 1;
		break;
	    case 'l':
		lock = 1;
		break;
	    case 'L':
		lock = 0;
		break;
	    case 'f':
		if (!strcmp(optarg, "turtle")) {
		    format = TS_TURTLE;
		} else if (!strcmp(optarg, "rdfxml")) {
		    format = TS_XML;
		} else if (!strcmp(optarg, "ntriples")) {
		    format = TS_NTRIPLES;
		} else if (!strcmp(optarg, "auto")) {
		    format = TS_AUTO;
		} else {
		    fprintf(stderr, "Unknown format: %s\n", optarg);
		    exit(1);
		}
		break;
	    case 1:
		if (!con) {
		    con = tsp_connect(NULL);
		    con->lock_tables = lock;
		    if (bulk) {
			ts_bulk_import_init(con, verbosity, stdout);
		    }
		}
		uri = ts_path_to_uri(optarg);
		if (!model) {
		    cmodel = uri;
		} else {
		    cmodel = model;
		}
		if (verbosity > 0) {
		    printf("import <%s>\n", cmodel);
		    /* flush stdout incase were running inside nohup or similar */
		    fflush(stdout);
		}
		if (!strcmp(TS_SYS_MODEL, cmodel)) {
		    fprintf(stderr, "Loading data into the system model is inadvisable, refusing\n");
		    break;
		}
		if (!add && !bulk) {
		    /* delay the cleaning of the models table - we will
		     * probably put back many of the same triples, and
		     * can do it once for all the imports */
		    erased += ts_erase(con, cmodel, TS_MODEL_DATA |
			            TS_MODEL_DELAY_CLEANUP, verbosity, stdout);
		}
		if (bulk) {
		    int num = ts_bulk_import(con, uri, cmodel, verbosity, stdout, format, optind == argc);
		    if (num > 0) {
			count += num;
		    }
		    if (verbosity > 1) {
			gettimeofday(&now, 0);
			diff = now.tv_sec - then.tv_sec +
			     (now.tv_usec - then.tv_usec) * 0.000001;
			printf("[S] %.1f\t%d\n", diff, count);
			/* flush stdout incase were running inside nohup or similar */
			fflush(stdout);
		    }
		} else {
		    int num = ts_import(con, uri, cmodel, verbosity, stdout, format);
		    if (num > 0) {
			count += num;
		    }
		}
		free(uri);
		done++;
		break;
            default:
                help++;
                break;
            }
        }
    }

    /* cleanup any crap left by TS_MODEL_DELAY_CLEANUP */
    if (erased > 0) {
	ts_orphan_cleanup(con);
    }
    if (bulk) {
	if (verbosity > 1) {
	    int l1, l2, miss;

	    ts_bulk_import_get_cache_stats(&l1, &l2, &miss);
	    printf("symbol cache stats: L1 hits: %d, L2 hits: %d, misses: %d\n",
		   l1, l2, miss);
	}
	ts_bulk_import_commit(con, verbosity, stdout);
	if (verbosity > 1) {
	    gettimeofday(&now, 0);
	    diff = now.tv_sec - then.tv_sec +
		 (now.tv_usec - then.tv_usec) * 0.000001;
	    printf("[S] %.1f\t%d\n", diff, count);
	}
    }

    if (con) {
	ts_disconnect(con);
    }
    ts_finish();

    if (help || !done) {
        fprintf(stderr, "Usage: %s "TSP_OPTIONS" [-v] [-m model] [-f format] [-b] <URI>+\n" TSP_HELP, basename(argv[0]));
	fprintf(stderr, "   -b, --background use background graph\n");
	fprintf(stderr, "   -a, --add       add document to model\n");
	fprintf(stderr, "   -l, --lock      lock tables during import%s\n",
		DEFAULT_LOCK_TABLES ? " (default)" : "");
	fprintf(stderr, "   -L, --no-lock   don't lock tables during import%s\n",
		DEFAULT_LOCK_TABLES ? " (default)" : "");
	fprintf(stderr, "   -m, --model     use model URI for import (default file's URI)\n");
	fprintf(stderr, "   -f, --format    select format for RDF file (default auto)\n");
	fprintf(stderr, "      where format is one of auto, rdfxml, turtle, ntriples\n");
	fprintf(stderr, "See man pages for more details\n");

        return 1;
    }

    TIME(TS_TIMING_REPORT);
    if (verbosity > 0) {
	gettimeofday(&now, 0);
	diff = now.tv_sec - then.tv_sec +
		     (now.tv_usec - then.tv_usec) * 0.000001;
	printf("loaded %d triples at %d triples/sec\n", count, (int)((double)count / diff));
    }

    return 0;
}

/* vi:set ts=8 sts=4 sw=4: */
